import simplejson as json
import logging
import os
import re
import pathlib
import string
import uuid
from typing import Dict, Optional

import boto3
import jwt
import requests
from urllib3.exceptions import InsecureRequestWarning

from ..direct._exceptions import ForbiddenError
from ._delivery_config import DeliveryConfig, DeliveryType
from ..direct._config import _Config

from ..direct._backend_apis import DeliveryAPIBackend
from ..direct._exceptions import BadRequestException, UnauthorizedDeliveryException

# NOTE: We are not able to verify the SSL cert on MDS API since
#       the VPC endpoint domain name is not on the server's SSL cert.
#       So with a heavy heart, we disable the InsecureRequestWarning
requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)


_logger = logging.getLogger(__name__)
_config = _Config()

# A filename can include 'a-z', 'A-Z', '0-9', '-', '_', '(', ')', ' ', '.'
_invalid_chars_regex = re.compile(r"[^a-zA-Z\d\-_\(\)\ .]")


def _get_user_id_from_uim_token() -> str:
    """Returns the user's id by decoding their UIM token"""
    try:
        md_auth_token = md_auth_token = os.getenv("MD_AUTH_TOKEN", os.getenv("UIM_JWT", ""))
        decode_key = os.getenv("UIM_DECODE_KEY", "https://login-prod.morningstar.com/.well-known/jwks.json")
        jwks_client = jwt.jwks_client.PyJWKClient(decode_key)
        signing_key = jwks_client.get_signing_key_from_jwt(md_auth_token)
        decoded_jwt = jwt.decode(
            md_auth_token,
            signing_key.key,
            algorithms=["RS256"],
            options={"verify_signature": True, "verify_exp": True, "verify_aud": False},
            leeway=60,
        )
        user_id: str = decoded_jwt["sub"].split("|")[1]
        return user_id.upper()
    except Exception as err:
        _logger.error(f"Error decoding MD_AUTH_TOKEN: {err}")
        raise


def _get_user_id_from_eams_token() -> str:
    """Returns the user's id by decoding their EAMS token"""
    try:
        key = os.getenv("EAMS_PUBLIC_KEY", "").replace("\\n", "\n")
        token = os.getenv("EAMS_JWT")
        decoded = jwt.decode(token, key, algorithms=["RS256"])
        user_id: str = decoded["sub"]
        return user_id.upper()
    except Exception as err:
        _logger.error(f"Error decoding EAMS token: {err}")
        raise


def _get_user_id() -> str:
    """Returns the user's id using their UIM token if it exists, EAMS token otherwise"""
    try:
        user_id = _get_user_id_from_uim_token()
        _logger.info(f"Using UIM token for User ID: {user_id}")
        return user_id
    except Exception:
        pass

    user_id = _get_user_id_from_eams_token()
    _logger.info(f"Using EAMS token for User ID: {user_id}")
    return user_id


def _get_job_name() -> str:
    """Returns the job name for a feed or delivery"""
    return str(uuid.uuid4())


def _get_request_id() -> str:
    """Returns the request id for use in POST requests"""
    request_id = os.getenv("REQUEST_ID", str(uuid.uuid4()))
    _logger.info(f"RequestId: {request_id} passed to POST requests")
    return request_id


def _get_bucket_name() -> str:
    """Returns the mds bucket name for s3 upload"""
    return os.getenv("MDS_BUCKET_NAME", "velo-notebook-output-prod-us-east-1")


def _create_feed_body() -> Dict:
    """Creates the post body which is used to call the MDS api when creating a feed id"""
    body = {
        "userId": _get_user_id(),
        "userName": _get_user_id(),
        "jobName": _get_job_name(),
        "jobMetaData": {
            "jobRunUrl": "place_holder",
            "jobRunMethod": "POST",
            "jobPostBody": {
                "app_source": "Morningstar Data Service(MDS)",
                "run_type": "notebook",
                "notebook_parameters": {"bucket": _get_bucket_name()},
            },
        },
    }
    return body


def _get_mds_base_url() -> str:
    return f"{_config.mds_api_url()}"


def _get_al_proxy_base_url() -> str:
    return f"{_config.al_proxy_api_url()}"


def _encrypt_ftp_password(password: str) -> str:
    """Takes a password and encrypts it using the user_id as the salt"""
    _api_backend = DeliveryAPIBackend()
    url = f"{_get_al_proxy_base_url()}encrypt-message"
    data = {"message": password, "salt": _get_user_id()}
    try:
        _logger.info(f"Calling {url} to encrypt ftp password")
        response_json = _api_backend.do_post_request(url, data=json.dumps(data, ignore_nan=True))
        encrypted_password: str = response_json["encrypted_message"]
        return encrypted_password
    except Exception:
        _logger.error("Could not encrypt password using /encrypt-message")
        raise


def _create_deliver_body(output_file_name: str, config: DeliveryConfig, delivered_file_name: Optional[str] = None) -> Dict:
    """Creates the post body which is used to call the MDS delivery api"""
    body = {
        "userId": _get_user_id(),
        "userName": _get_user_id(),
        "jobName": _get_job_name(),
        "s3key": output_file_name,
    }
    if delivered_file_name:
        body["fileName"] = delivered_file_name
    if config.method == DeliveryType.EMAIL:
        body["deliveryInfos"] = {  # type: ignore
            "emails": [
                {
                    "address": config.email_address,
                },
            ]
        }
    elif config.method == DeliveryType.DELIVERY_PROFILE:
        body["deliveryInfos"] = {  # type: ignore
            "profiles": [
                {
                    "deliveryProfileId": config.delivery_profile_id,
                },
            ]
        }
    elif config.method == DeliveryType.FTP:
        body["deliveryInfos"] = {  # type: ignore
            "ftps": [
                {
                    "server": config.server,
                    "username": config.user_name,
                    "password": _encrypt_ftp_password(config.password),
                    "folder": config.folder,
                },
            ]
        }
    return body


def _create_output_file_name(file_path: str) -> str:
    """Creates the output file name for the file which will be uploaded to MDS s3 bucket"""
    file_name = pathlib.Path(file_path).name
    return f"{uuid.uuid4()}-{file_name}"


def _retrieve_credentials() -> Dict[str, str]:
    """Retrieves the credentials required to access the MDS s3 bucket"""
    _api_backend = DeliveryAPIBackend()
    url = f"{_get_al_proxy_base_url()}assume-mds-role?user_id={_get_user_id()}"
    try:
        _logger.info(f"Calling {url} to get assume role credentials")
        response_json: Dict[str, str] = _api_backend.do_get_request(url)
        return response_json
    except ForbiddenError:
        _logger.error(f"Unable to retrieve credentials for MDS s3 bucket upload")
        _logger.error(f"User does not have access to deliver function feature")
        raise UnauthorizedDeliveryException from None


def _get_mds_bucket() -> boto3.session.Session.resource:
    """Returns the MDS bucket where we will upload the file before calling the delivery endpoint"""
    credentials = _retrieve_credentials()
    s3_resource = boto3.resource(
        "s3",
        aws_access_key_id=credentials["AccessKeyId"],
        aws_secret_access_key=credentials["SecretAccessKey"],
        aws_session_token=credentials["SessionToken"],
    )
    bucket = s3_resource.Bucket(_get_bucket_name())
    return bucket


def _validate_file_name(file_name: str) -> bool:
    """Checks if the file name used for the delivered file is valid"""
    matches = _invalid_chars_regex.findall(file_name)
    if matches:
        error_message = f"File name: {file_name} is invalid. The following can't be used in the file name: '{matches}'"
        raise BadRequestException(error_message)
    return True



import simplejson as json
from abc import ABC, abstractmethod
from simplejson import JSONEncoder
from typing import Any, Optional

from .._error_messages import RESOURCE_NOT_FOUND_ERROR_PORTFOLIO_List
from ..._base import _logger
from .. import _error_messages
from .._base_api import APIBackend


class PortfolioJSONEncoder(JSONEncoder):
    def default(self, obj: Any) -> Any:
        return obj.to_json()


class PortfolioNotFoundError(Exception):
    def __init__(self, message: Optional[str] = None) -> None:
        if not message:
            super().__init__(_error_messages.RESOURCE_NOT_FOUND_ERROR)
        else:
            super().__init__(message)


class PortfolioDataApiBackend(APIBackend):
    """
    Subclass to call the Portfolio Data API and handle any HTTP errors that occur.
    """

    def __init__(self) -> None:
        super().__init__()

    def _handle_custom_http_errors(self, res: Any) -> Any:
        """
        Handle HTTP errors with custom error messages
        """
        response_message = res.json()["message"]
        if res.status_code == 404:
            _logger.debug(f"Resource Not Found Error: {res.status_code} {response_message}")
            raise PortfolioNotFoundError(RESOURCE_NOT_FOUND_ERROR_PORTFOLIO_List) from None



from enum import Enum

import pandas
from pandas import DataFrame
from typing import Optional, Any, List
import re

from . import _decorator, _error_messages
from .._base import _logger
from ._exceptions import ValueErrorException, ResourceNotFoundError, BadRequestException
from ..direct._config import _Config
from .data_type import Universe
from ._data_point import _get_data_point_details, _get_asset_flow_data_points_by_ids, _get_all_universes
from ._utils import (
    _empty_to_none,
    _extract_data,
    _filter_data_frame_column_by_setting,
    _reindex_data_frame_column,
)
from ._config_key import ALL_ASSET_FLOW_DATA_POINTS
from ._base_api import APIBackend
from urllib.parse import quote

_config = _Config()


class _LookupModule(Enum):
    DATA_SET = 0
    DATA_POINT = 1
    INVESTMENT = 2


class LookupDataAPIBackend(APIBackend):
    """
    Subclass to call the Lookup Data API and handle any HTTP errors that occur.
    """

    def __init__(self, lookup_module: _LookupModule) -> None:
        super().__init__()
        self._lookup_module = lookup_module

    def _get_resource_not_found_message(self) -> str:
        if self._lookup_module == _LookupModule.DATA_SET:
            return _error_messages.RESOURCE_NOT_FOUND_ERROR_DATA_SET_LOOKUP
        elif self._lookup_module == _LookupModule.DATA_POINT:
            return _error_messages.RESOURCE_NOT_FOUND_ERROR_DATA_POINT_LOOKUP
        return _error_messages.RESOURCE_NOT_FOUND_ERROR

    def _handle_custom_http_errors(self, res: Any) -> Any:
        """
        Handle HTTP errors with custom error messages
        """
        response_message = res.json()["message"]
        if res.status_code == 404:
            _logger.debug(f"Resource Not Found Error: {res.status_code} {response_message}")
            raise ResourceNotFoundError(self._get_resource_not_found_message()) from None


_data_set_lookup_api_request = LookupDataAPIBackend(_LookupModule.DATA_SET)
_investment_lookup_api_request = LookupDataAPIBackend(_LookupModule.INVESTMENT)

_data_point_lookup_api_request = LookupDataAPIBackend(_LookupModule.DATA_POINT)


class investment_universes:
    All_Managed_Investments = Universe("All Managed Investments", "MI")
    Bonds = Universe("Bonds", "BD")
    Closed_End_Funds = Universe("Closed End Funds", "FC")
    Collective_Investment_Trusts = Universe("Collective Investment Trusts", "CZ")
    European_Pension_Life_Fund_Wrappers = Universe("European Pension/Life Fund Wrappers", "PS")
    eVestment_Hedge_Funds = Universe("eVestment Hedge Funds", "VH")
    eVestment_Separate_Accounts = Universe("eVestment Separate Accounts", "VS")
    Exchange_Traded_Funds = Universe("Exchange-Traded Funds", "FE")
    Funds_Open_End_and_Exchange_Traded_Funds = Universe("Funds (Open End and Exchange-Traded Funds)", "FX")
    France_Dedicated_Funds = Universe("France Dedicated Funds", "DF")
    Global_Restricted_Funds = Universe("Global Restricted Funds", "R1")
    Hedge_Funds = Universe("Hedge Funds", "FH")
    HFR_Hedge_Funds = Universe("HFR Hedge Funds", "H1")
    HSBC_Restricted_Funds = Universe("HSBC Restricted Funds", "IF")
    Insurance_and_Pension_Funds = Universe("Insurance and Pension Funds", "FV")
    Market_Indexes = Universe("Market Indexes", "XI")
    Models = Universe("Models", "MO")
    Money_Market_Funds = Universe("Money Market Funds", "FM")
    Open_End_Funds = Universe("Open End Funds", "FO")
    Plans_529 = Universe("529 Plans", "CP")
    Portfolios_529 = Universe("529 Portfolios", "CT")
    Private_Funds = Universe("Private Funds", "SP")
    Separate_Accounts = Universe("Separate Accounts", "SA")
    Stocks = Universe("Stocks", "ST")
    Strategies = Universe("Strategies", "MG")
    UBS_Separate_Accounts = Universe("UBS Separate Accounts", "S1")
    UK_Life_and_Pension_Funds = Universe("UK Life and Pension Funds", "V1")
    Unit_Investment_Trust = Universe("Unit Investment Trust", "FI")
    US_Variable_Annuities = Universe("US Variable Annuities", "VP")
    US_Variable_Annuity_Subaccounts = Universe("US Variable Annuity Subaccounts", "VA")
    US_Variable_Life = Universe("US Variable Life", "LP")
    US_Variable_Life_Subaccounts = Universe("US Variable Life Subaccounts", "VL")


def _search_data(datapoint: str, keyword: str, universe: str, findBy: str = "name") -> Any:
    # Encode keyword.
    _logger.debug(f"Encode the {keyword} keyword in utf-8")
    keyword_encode = quote(keyword, "utf-8")

    _logger.debug(f"Search by {findBy} in the {datapoint} data point with: keyword: {keyword}, universe: {universe}")
    url = f"{_config.securitydata_service_url()}v1/find?datapoint={datapoint}&kw={keyword_encode}&operator=namecontain&universe={universe}&findby={findBy}&status=activeonly"

    response_json = _investment_lookup_api_request.do_get_request(url)
    return response_json


@_decorator.typechecked
def get_morningstar_data_sets(universe: Optional[str] = None) -> DataFrame:
    """Returns all Morningstar pre-defined data sets.

    Args:
        universe (:obj:`str`, `optional`): Investment universe code. Example: "FO". Use `get_investment_universes <./lookup.html#morningstar_data.direct.lookup.get_investment_universes>`_ to discover possible values.

    Returns:
        DataFrame: A DataFrame object with Morningstar data sets. DataFrame columns include:

        * datasetId
        * name

    Examples:
        Retrieve the Morningstar data set for the open-end fund universe.

    ::

        import morningstar_data as md

        df = md.direct.lookup.get_morningstar_data_sets(universe="FO")
        df

    :Output:
        =========  ===============================
        datasetId  name
        =========  ===============================
        0026-0020  Snapshot
        0026-0447  Sustainability: ESG Risk (Fund)
        ...
        =========  ===============================

    Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    if universe:
        url = f"{_config.data_point_service_url()}v1/defaultviews?type=screen&universe={universe}"
    else:
        url = f"{_config.data_point_service_url()}v1/defaultviews?type=list"
    response_json = _data_set_lookup_api_request.do_get_request(url)
    result = DataFrame(response_json)
    result["datasetId"] = result["id"]
    return result[["datasetId", "name"]]


def _get_multiple_data_point_details(params: list) -> list:
    asset_flow_data_points = []
    normal_data_points = []
    params = params if params else []
    for data_point in params:
        if data_point is None:
            continue
        if data_point.get("datapointId", "").strip() in ALL_ASSET_FLOW_DATA_POINTS:
            asset_flow_data_points.append(data_point.get("datapointId", "").strip())
        else:
            normal_data_points.append(data_point)

    result = []
    if normal_data_points:
        normal_response_json = _get_data_point_details(normal_data_points)
        if isinstance(normal_response_json, list) and len(normal_response_json) > 0:
            result.extend(normal_response_json)

    if asset_flow_data_points:
        result.extend(_get_asset_flow_data_points_by_ids(asset_flow_data_points))

    return _data_points_reorder(params, result)


def _data_points_reorder(params: list, result: list) -> list:
    target = []
    for data_point in params:
        target.extend(_data_points_mapping(data_point, result))
    return target


def _data_points_mapping(data_point: Optional[dict], result: list) -> List[Any]:
    mapping_result: List[Any] = []
    if data_point is None:
        return mapping_result
    data_point_id = data_point.get("datapointId", "").strip()
    is_ts_dp = data_point.get("isTsdp", None)
    if data_point_id in ALL_ASSET_FLOW_DATA_POINTS:
        target_list = list(filter(lambda x: (data_point_id == x.get("datapointId", "").strip()), result))
        if len(target_list) > 0:
            mapping_result.append(target_list[0])
    elif is_ts_dp is not None:
        target_list = list(
            filter(
                lambda x: (data_point_id == x.get("datapointId", "").strip()) & (is_ts_dp == x.get("isTsdp", None)),
                result,
            )
        )
        if len(target_list) > 0:
            mapping_result.append(target_list[0])
    else:
        target_list_for_ts = list(
            filter(
                lambda x: (data_point_id == x.get("datapointId", "").strip()) & (True == x.get("isTsdp", None)),
                result,
            )
        )
        target_list_for_no_ts = list(
            filter(
                lambda x: (data_point_id == x.get("datapointId", "").strip()) & (False == x.get("isTsdp", None)),
                result,
            )
        )
        if len(target_list_for_ts) > 0:
            mapping_result.append(target_list_for_ts[0])
        if len(target_list_for_no_ts) > 0:
            mapping_result.append(target_list_for_no_ts[0])

    return mapping_result


@_decorator.not_null
@_decorator.typechecked
def get_data_point_settings(data_point_ids: List[str]) -> DataFrame:
    """Returns settings for a given set of data points. This settings DataFrame can be manipulated
    to reflect specific settings to be used for data retrieval.

    Args:
        data_point_ids (:obj:`list`): A list of unique identifiers for data points.
            Example: ["OS01W", "HP010"]

    Returns:
        DataFrame: A DataFrame object with data point settings. DataFrame columns include:

        * datapointId
        * datapointName
        * displayName
        * currency
        * preEuroConversion
        * sourceId
        * frequency
        * startDate
        * endDate
        * floatStart
        * floatEnd
        * startDelay
        * endDelay
        * diffStart
        * diffEnd
        * compounding
        * calculationId
        * annualized
        * annualDays
        * benchmark
        * riskfree
        * windowType
        * windowSize
        * stepSize
        * requireContinueData
        * fit
        * scalType
        * scalValue
        * scalPercentValue
        * timehorizon

    Examples:
        Get data point settings for data point "OS01W".

    ::

        import morningstar_data as md

        df = md.direct.get_data_point_settings(data_point_ids=["OS01W"])
        df

    :Output:
        ===========  =============  ===  ========  ========  =========
        datapointId  datapointName  ...  calcMnav  showType  transType
        ===========  =============  ===  ========  ========  =========
        OS01W	     Name           ...  None      None      None
        ===========  =============  ===  ========  ========  =========

    Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    params = list(map(lambda x: {"datapointId": x}, data_point_ids))
    response_json = _get_multiple_data_point_details(params)
    if len(response_json) == 0:
        raise ValueErrorException("There are no data points.")
    extracted_data = _extract_data(response_json)
    details = DataFrame(extracted_data)
    filtered_details = _filter_data_frame_column_by_setting(details)
    details_with_relocated_data_point_id = _reindex_data_frame_column(filtered_details, target_column="datapointId", new_index=0)
    details_with_relocated_data_point_name = _reindex_data_frame_column(details_with_relocated_data_point_id, target_column="datapointName", new_index=1)
    return _empty_to_none(details_with_relocated_data_point_name)


@_decorator.not_null
@_decorator.typechecked
def investments(keyword: str, count: int = 50, only_surviving: bool = True) -> DataFrame:
    """Returns investments that match the given keyword.

    Args:
        keyword (:obj:`str`): Keyword to search for investments.
        count (:obj:`int`, `optional`): Maximum number of investments to return that match the keyword.
        only_surviving (:obj:`bool`, `optional`): Include only surviving investments.
    Returns:
        DataFrame: A DataFrame object with investments. DataFrame columns include:

        * Name
        * SecId
        * Ticker
        * ISIN
        * CUSIP
        * Base Currency
        * Exchange
        * Country
        * Security Type
        * Fund Id
        * Performance Id


    Examples:
        Get investments that match the keyword "fixed".

    ::

        import morningstar_data as md

        df = md.direct.lookup.investments(keyword="fixed")
        df

    :Output:
        ================================= ============= === ============== ============ ==================
        Name                                     SecId  ...  Security Type      Fund Id     Performance Id
        ================================= ============= === ============== ============ ==================
        Fixed.zone AS	                    0P0001PHEP  ...             ST                      0P0001PHEP
        Fixed Income Allocation CL I1       FS0000G0QH  ...             CZ   FS0000G0QH         0P0001JW76
        ================================= ============= === ============== ============ ==================

    Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """

    # Encode keyword.
    keyword_encode = quote(keyword, "utf-8")
    _logger.debug(f"Search investments with:{keyword_encode}, Top {count} of investments to return.")
    url = f"{_config.securitydata_service_url()}v1/autocomplete?checkEntitlement=true&count={count}&keyword={keyword_encode}&onlySurviving={str(only_surviving).lower()}"
    response_json = _investment_lookup_api_request.do_get_request(url)
    investments = response_json["investments"]

    _logger.info(f"No investments return with keyword:{keyword}")
    if not investments:
        _logger.debug(f"No investments return with keyword:{keyword}")
        return DataFrame({"Name": [], "SecId": [], "ISIN": [], "CUSIP": [], "Base Currency": []})

    result = DataFrame(investments)
    # Remove some unused data.
    _logger.info(f"Remove some unused data")
    _logger.debug(f"{len(investments)} investments return. total investments found:{response_json['total']}")
    new_data_frame = result.drop(columns=["targetUniverseId", "csdccCode", "score", "itaCode"])
    # Rename and reoreder columns.
    _logger.info(f"Rename and reoreder columns. data")
    new_data_frame = new_data_frame.rename(
        columns={
            "name": "Name",
            "secId": "SecId",
            "ticker": "Ticker",
            "isin": "ISIN",
            "baseCurrency": "Base Currency",
            "cusip": "CUSIP",
            "exchange": "Exchange",
            "countryOrientation": "Country",
            "securityType": "Security Type",
            "fundId": "Fund Id",
            "performanceId": "Performance Id",
        }
    )
    return new_data_frame[["Name", "SecId", "Ticker", "ISIN", "CUSIP", "Base Currency", "Exchange", "Country", "Security Type", "Fund Id", "Performance Id"]]


@_decorator.not_null
@_decorator.typechecked
def get_morningstar_category(universe: str) -> DataFrame:
    """Returns Morningstar Categories for the given universe.

    Args:
        universe (:obj:`str`): Investment universe code. Example: "FO". Use `get_investment_universes <./lookup.html#morningstar_data.direct.lookup.get_investment_universes>`_ to discover possible values.

    Returns:
        DataFrame: A DataFrame object with Morningstar Category details. DataFrame columns include:

        * Morningstar Category
        * Category Code
        * Global Broad Category Id
        * Global Broad Category Group
        * Region

    Examples:
        Retrieve the Morningstar Category for the open-end fund universe.

    ::

        import morningstar_data as md

        df = md.direct.lookup.get_morningstar_category("FO")
        df

    :Output:
        ==========================  =============  ========================  ===========================  ==========
        Morningstar Category        Category Code  Global Broad Category Id  Global Broad Category Group  Region
        ==========================  =============  ========================  ===========================  ==========
        10 yr Government Bond	    INCA000008	   $BCG$FXINC	             Fixed Income	              India
        2025 Target Date Portfolio	CACA000158	   $BCG$ALLOC	             Allocation	                  Canada
        2030 Target Date Portfolio	CACA000164	   $BCG$ALLOC	             Allocation	                  Canada
        ...
        ==========================  =============  ========================  ===========================  ==========
    Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    url = f"{_config.data_point_service_url()}v1/datapoints/datachoiceoptions?universe={universe}"

    _logger.debug(f"Lookup Morningstar Category from DataPoint API for universe: {universe}")

    # OF003 is datapoint 'Mornignstar Category' which identifies funds based on their actual investment styles as measured by their underlying portfolio holdings
    res = _data_point_lookup_api_request.do_post_request(url, "OF003")

    if not res or not res[0].get("options", []):
        return DataFrame(columns=["Morningstar Category", "Category Code", "Global Broad Category Id", "Global Broad Category Group", "Region"])

    options = res[0].get("options", [])

    data = [
        {
            "Morningstar Category": x.get("name", ""),
            "Category Code": x.get("value", ""),
            "Global Broad Category Id": x.get("additional", {}).get("globalBroadCategoryId", ""),
            "Global Broad Category Group": x.get("additional", {}).get("globalBroadCategoryName", ""),
            "Region": x.get("additional", {}).get("regionName", ""),
        }
        for x in options
    ]

    return DataFrame(data)


@_decorator.not_null
@_decorator.typechecked
def firms(keyword: str, universe: Optional[str] = None) -> DataFrame:
    """Returns firms that match the given keyword.

    Args:
        keyword (:obj:`str`): Keyword to search for firms.
        universe (:obj:`str`, `optional`): Investment universe code. Example: "FO". Use `get_investment_universes <./lookup.html#morningstar_data.direct.lookup.get_investment_universes>`_ to discover possible values.

    Returns:
        DataFrame: A DataFrame object with firms. DataFrame columns include:

            * Id
            * Firm Name
            * Universe


    Examples:
        Get all firms that match the keyword "rock".

    ::

        import morningstar_data as md

        df = md.direct.lookup.firms("rock")
        df

    :Output:
        =============  =======================================  =============
        Id              Firm Name                                  Universe
        =============  =======================================  =============
        F00000Z17E      Amrock Capital B.V.                             FO
        F00000XC6D      BlackRock	                                    FO
        F00000SEJ6      BlackRock (Channel Islands) Limited	            FO
        ...
        =============  =======================================  =============

    Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    # Encode keyword.
    keyword_encode = quote(keyword, "utf-8")
    # Encode universe.
    if universe:
        univ_encode = quote(universe.upper(), "utf-8")
    else:
        univ_encode = ""
    _logger.debug(f"Search firms with:{keyword} and universe:{universe} ")
    # OF006 is Datapoint 'Firm Name', A company which offers mutual funds.
    url = f"{_config.securitydata_service_url()}v1/find?datapoint=OF006&findby=namecontain&kw={keyword_encode}&universe={univ_encode}"

    response_json = _investment_lookup_api_request.do_get_request(url)

    fields = ["name", "additional.universe"]
    firms_df = pandas.json_normalize(response_json)
    if firms_df.empty:
        _logger.debug(f"Search firms without result. keyword:{keyword} and universe:{universe} ")
        return DataFrame({"Id": [], "Name": [], "Universe": []})
    firms_df = firms_df.rename(columns={"name": "Name", "additional.universe": "Universe", "id": "Id"})
    return firms_df


@_decorator.typechecked
def get_brandings(keyword: str, find_by: str, universe: Optional[str] = None) -> DataFrame:
    """Returns brandings for the given universe and keyword.

    Args:
        keyword(:obj:`str`): Keyword to search for brandings.
        find_by(:obj:`str`): Search condition to search for brandings. Valid values are ``name_begin`` and ``name_contain``.
        universe(:obj:`str`, `optional`): Investment universe code. Example: "FO". Use `get_investment_universes <./lookup.html#morningstar_data.direct.lookup.get_investment_universes>`_ to discover possible values.

    Returns:
        DataFrame: A DataFrame object with brandings. DataFrame columns include:

        * Id
        * Name

    Examples:
        Search brandings in the open-end fund universe, where branding name begins with the keyword "a".

    ::

        import morningstar_data as md

        df = direct.lookup.get_brandings("a", "name_begin", "FO")
        df

    :Output:
        ==========  ===============================
        Id            Name
        ==========  ===============================
        BN000007Q5    A Plus Finance
        BN000007Q6    A&G
        BN00000GBI    A1
        ...
        ==========  ===============================

    Errors:
        BadRequestException: Raised when the parameter find_by is not a valid value.

        AccessDeniedError: Raised when the user is not authenticated.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.
    """

    if find_by not in ["name_begin", "name_contain"]:
        raise BadRequestException(message="Valid values for find_by should be `name_begin` or `name_contain`.")

    condition = find_by.replace("_", "")

    # FC001 is a global grouping of asset management companies that represents the management philosophy,
    # the firm's marketing/communication channels, and their distribution/sales efforts.
    url = f"{_config.securitydata_service_url()}v1/find?datapoint=FC001&findby={condition}&kw={keyword}&universe={universe}"

    res = _investment_lookup_api_request.do_get_request(url)

    if not res:
        return DataFrame(columns=["Id", "Name"])

    data = [{"Id": x.get("id", ""), "Name": x.get("name", "")} for x in res]

    return DataFrame(data)


@_decorator.typechecked
def portfolio_managers(keyword: str = "", universe: str = "FO") -> DataFrame:
    """
    Returns portfolio managers for the given universe and keyword.

    Args:
        keyword (:obj:`str`): String to match one or more manager names. If the keyword is empty or not passed to the function, the function will return all managers.

        universe (:obj:`str`): Investment universe code. Example: "FO". Use `get_investment_universes <./lookup.html#morningstar_data.direct.lookup.get_investment_universes>`_ to discover possible values.


    Returns:
        DataFrame: A DataFrame object with portfolio managers. DataFrame columns include:

            * Id
            * Manager Name
            * Firm

    Examples:
        Get portfolio managers that match the keyword "alla".

    ::

        import morningstar_data as md

        df = md.direct.lookup.portfolio_managers(keyword="alla")
        df

    Output:
        ======  =======================  ===============
        Id   	Manager name	         Firm
        ======  =======================  ===============
        81643	Abdallah Guezour	     Schroders
        123653	Abdallah Nauphal	     TD
        213243	Adam Rizkalla	         River Canyon
        153647	Adrian Allardice	     Old Mutual
        182699	Adrian van Pallander	 Coronation
        ...
        ======  =======================  ===============

    Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    # FCM01 is Datapoint which name is 'Manager & Fund Company'
    _logger.debug(f"Search managers with: keyword: {keyword}, universe: {universe} in the FCM01 datapoint")
    data = _search_data("FCM01", keyword, universe)

    _logger.info(f"Total managers found: {len(data)} managers")
    result = DataFrame(data)

    _logger.debug("Rename columns")
    result.rename(columns={"id": "Id", "name": "Manager name"}, inplace=True)

    if result.empty:
        _logger.info(f"No managers return with keyword: {keyword}")
        return DataFrame({"Id": [], "Manager name": [], "Firm": []})

    _logger.debug("Select only the Id and Manager name in the DataFrame")
    portfolio_managers_result = result[["Id", "Manager name"]]

    _logger.debug("Apply the _separate_name_and_firm method in the columns to separate the manager name and the firm")
    portfolio_managers_result = portfolio_managers_result.apply(_separate_name_and_firm, axis=1)

    return portfolio_managers_result


# The name column has the manager's name and the firm name in the same column. This function looks for separating them into different columns
# INPUT: name column: "Abdallah Guezour (Schroders)"
# OUTPUT: name column: "Abdallah Guezour | firm column: "Schroders"
def _separate_name_and_firm(portfolio_manager: DataFrame) -> DataFrame:
    manager_name = portfolio_manager["Manager name"]
    # Check if the name does not contain the firm within (). If don't, the firm value will be None
    if manager_name.count("(") < 1:
        portfolio_manager["Firm"] = None
    # If the name contains the firm name within "()" we will separate them in different columns
    else:
        # Extract the firm name with an regex expression based on the name value
        portfolio_manager["Firm"] = re.findall("\((.+?)\)", manager_name)[0]
        # Remove signature name from column name value
        portfolio_manager["Manager name"] = manager_name.replace("(" + portfolio_manager["Firm"] + ")", "").strip()

    return portfolio_manager


@_decorator.typechecked
def companies(keyword: str = "", universe: str = "ST") -> DataFrame:
    """
    Returns companies for the given universe and keyword.

    Args:
        keyword (:obj:`str`): String to match one or more company names. If the keyword is empty or not passed to the function, the function will return all companies.

        universe (:obj:`str`): Investment universe code. Example: "FO". Use `get_investment_universes <./lookup.html#morningstar_data.direct.lookup.get_investment_universes>`_ to discover possible values.


    Returns:
        DataFrame: A DataFrame object with companies. DataFrame columns include:

            * Id
            * Company name

    Examples:
        Get companies that match the keyword "Energy".

    ::

        import morningstar_data as md

        df = md.direct.lookup.companies("Energy")
        df

    :Output:
        ==========  ========================
        Id	        Company name
        ==========  ========================
        0C00008O2A	11 Good Energy Inc
        0C000017RQ	2G Energy AG
        0C0000516I	3MV Energy Corp
        0C00000ZAZ	3Power Energy Group Inc
        0C00000QLB	3TEC Energy Corp
        ...
        ==========  ========================

    Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    # AA0A5 is Datapoint which name is 'Company Name'
    _logger.debug(f"Search companies with: keyword: {keyword}, universe: {universe} in the AA0A5 datapoint")
    data = _search_data("AA0A5", keyword, universe)

    _logger.info(f"Total companies found: {len(data)} companies")
    result = DataFrame(data)

    _logger.debug("Rename columns")
    result.rename(columns={"id": "Id", "name": "Company name"}, inplace=True)

    if result.empty:
        _logger.info(f"No companies return with keyword: {keyword}")
        return DataFrame({"Id": [], "Company name": []})

    _logger.debug("Select only the Id and Company name in the DataFrame")
    companies_result = result[["Id", "Company name"]]

    return companies_result


@_decorator.typechecked
def get_investment_universes(include_category_universes: bool = False) -> DataFrame:
    """Returns investment universe names and IDs.
    For example, Bonds, eVestment Hedge Funds, Global Restricted Funds.

    Args:
        include_category_universes (:obj:`bool`, `optional`): If True, the function will return all universes, including category universes. Default is False.

    Returns:
        DataFrame: A DataFrame object with universes.  DataFrame columns include:

        * Id
        * Name

    :Examples:

    ::

        import morningstar_data as md

        df = md.direct.lookup.get_investment_universes()
        df

    :Output:
        ==== ==============
        Id   Name
        ==== ==============
        FH   Hedge Funds
        XI   Market_Indexes
        ...
        ==== ==============
    """

    universes = _get_all_universes()

    universes = universes.rename(columns={"id": "Id", "name": "Name"})

    if not include_category_universes:
        universes = universes[~universes["Id"].str.contains("CA]")]  # Drops all category universes

    return universes



import base64
import simplejson as json
import os
import re
from os import environ

from dataclasses import dataclass, field
from typing import Dict
from ._exceptions import EAMSTokenDoesNotExistError, MalformedJWTError, CredentialsException


@dataclass
class _Config:
    def __init__(self) -> None:
        self._MD_API: str = os.getenv("MD_API", "https://www.us-api.morningstar.com/md-api/")

    def get_uim_token(self) -> str:
        token = os.getenv("MD_AUTH_TOKEN", os.getenv("UIM_JWT", ""))
        if token == "":
            raise CredentialsException("Please set up MD_AUTH_TOKEN environment variable to access the morningstar_data package")
        return self._filter_invalid_jwt(token)

    def get_headers(self, add_headers: Dict[str, str] = {}) -> Dict[str, str]:
        headers = {
            "X-API-ComponentId": os.getenv("DO_API_REQUEST_ORIGIN", "analyticslab"),
            "X-API-Sourceapp": os.getenv("DO_API_REQUEST_ORIGIN", "morningstar-data"),
            "X-API-ProductId": "Direct",
            "Content-Type": "application/json",
            **add_headers,
        }

        # This will allow us to make DO API requests on behalf of another user.
        # EAMS impersonation is going away. We are using this temporarily.
        if environ.get("USE_EAMS_IMPERSONATION"):
            # EAMS_JWT is always on the environment in Analytics Lab.
            # We only want to use it in DNA Lab api when USE_EAMS_IMPERSONATION is set to True
            if "EAMS_JWT" in environ:
                headers["X-API-JsonWebToken"] = environ["EAMS_JWT"]
            else:
                raise EAMSTokenDoesNotExistError("EAMS_JWT does not exist on the environment.")
        else:
            headers["authorization"] = f"Bearer {self.get_uim_token()}"

        return headers

    def object_service_url(self) -> str:
        return f"{self._MD_API}proxy_request/object_service/"

    def custom_data_points_service_url(self) -> str:
        return f"{self._MD_API}proxy_request/custom_data_points_service/"

    def securitydata_service_url(self) -> str:
        return f"{self._MD_API}proxy_request/security_data_service/"

    def searches_service_url(self) -> str:
        return f"{self._MD_API}proxy_request/searches_service/"

    def investment_data_service_url(self) -> str:
        return f"{self._MD_API}v0/investment/investment_data"

    def portfolio_service_url(self) -> str:
        return f"{self._MD_API}proxy_request/portfolio_service/"

    def performancereport_service_url(self) -> str:
        return f"{self._MD_API}proxy_request/performancereport_service/"

    def performancereport_export_service_url(self) -> str:
        return f"{self._MD_API}proxy_request/performancereport_export_service/"

    def dataset_service_url(self) -> str:
        return f"{self._MD_API}proxy_request/dataset_service/"

    def data_point_service_url(self) -> str:
        return f"{self._MD_API}proxy_request/data_point_service/"

    def asset_service_url(self) -> str:
        return f"{self._MD_API}proxy_request/asset_service/"

    def peergroup_service_url(self) -> str:
        return f"{self._MD_API}proxy_request/peergroup_service/"

    def fof_api_url(self) -> str:
        return f"{self._MD_API}proxy_request/fof_service/"

    def al_proxy_api_url(self) -> str:
        return f"{self._MD_API}proxy_request/al_proxy_api/"

    def mds_api_url(self) -> str:
        return f"{self._MD_API}proxy_request/mds_api/"

    def temp_table_url(self) -> str:
        return f"{self._MD_API}proxy_request/temp_table_service/"

    def query_api_url(self) -> str:
        return f"{self._MD_API}proxy_request/lakehouse_query_service/"

    def _filter_invalid_jwt(self, jwt: str) -> str:
        # Define the regular expression pattern to match a valid JWT
        pattern = r"^[A-Za-z0-9_-]*\.?[A-Za-z0-9_-]*\.?[A-Za-z0-9_-]*$"

        # Check if the JWT matches the valid pattern
        if not re.match(pattern, jwt):
            # Replace invalid characters with an empty string
            jwt = re.sub(r"[^\w\.-]", "", jwt)

        # jwt = jwt.strip()
        # # Use regex to replace
        # jwt = re.sub(pattern, "", jwt)

        # Check that the JWT has 3 parts
        if len(jwt.split(".")) != 3:
            raise MalformedJWTError

        # Split the JWT into header, payload, and signature
        header, payload, signature = jwt.split(".")

        # Filter out any invalid characters from the header and payload
        decoded_header = base64.urlsafe_b64decode(header + "=" * (4 - len(header) % 4)).decode("utf-8", "ignore")
        decoded_payload = base64.urlsafe_b64decode(payload + "=" * (4 - len(payload) % 4)).decode("utf-8", "ignore")

        # Check that the filtered header and payload are valid JSON objects
        try:
            json.loads(decoded_header)
            json.loads(decoded_payload)
        except ValueError:
            raise MalformedJWTError from None

        # If we got here, the JWT is valid and filtered
        return ".".join([header, payload, signature])



from .investment_lists import get_investment_list, get_investment_lists, save_investment_list
from .search_criteria import get_search_results, get_search_criteria_condition, get_search_criteria_conditions, get_search_criteria, get_search_criterias
from .data_set import get_data_set, get_data_set_details, get_data_sets
from .portfolio import get_portfolios, save_portfolio
from .._portfolio import PortfolioType



import functools
import requests

from uuid import UUID
from typing import Union, List, Any, Dict, Optional, Callable
from .._api import _direct_api_request
from .._config import _Config
from .._exceptions import BadRequestException, ResourceNotFoundError
from .. import _error_messages
from ..._base import _logger
from ..user_items.search_criteria import execute_search
from .._error_messages import (
    BAD_REQUEST_ERROR_INVALID_CATEGORY_ID,
    BAD_REQUEST_ERROR_NO_INVESTMENT_SOURCE,
)



_config = _Config()

SearchObjectType = Optional[Dict[str, Any]]
InvestmentsListType = Optional[List[str]]
UUIDOrInitType = Optional[str]
InvestmentType = Union[SearchObjectType, UUIDOrInitType, InvestmentsListType]
CategoryType = Optional[str]


class Investments:
    def __init__(self, investment_object: InvestmentType) -> None:
        self.investment_ids: InvestmentsListType = None
        self.search_criteria: UUIDOrInitType = None
        self.list_id: UUIDOrInitType = None
        self.search_criteria_object: SearchObjectType = None
        self.category_id: CategoryType = None
        self._parse_investment_object(investment_object)

    def _parse_investment_object(self, investment_object: InvestmentType) -> None:
        if isinstance(investment_object, list):
            self.investment_ids = investment_object
        elif isinstance(investment_object, dict):
            self.search_criteria_object = investment_object
        elif self._is_int(investment_object):
            self.search_criteria = investment_object
        elif self._is_guid(investment_object):
            self.list_id = investment_object
        elif isinstance(investment_object, str) and investment_object is not None and len(investment_object.strip()) > 0:
            self.category_id = investment_object
        else:
            raise ValueError("Invalid investment object")

    def _is_int(self, value: InvestmentType) -> bool:
        try:
            if not isinstance(value, str):
                raise ValueError("Invalid investment object: search criteria must be a string")
            int(value)
            return True
        except ValueError:
            return False

    def _is_guid(self, value: InvestmentType) -> bool:
        try:
            if not isinstance(value, str):
                raise ValueError("Invalid investment object: list ID must be a string")
            UUID(value)
            return True
        except ValueError:
            return False

    @functools.lru_cache()
    def get_investment_ids(self) -> list:
        id_list = []
        if self.investment_ids:
            id_list = self.investment_ids

        elif self.list_id is not None and len(self.list_id.strip()) > 0:
            id_list = self._get_investment_ids_by_list_id()
            if not id_list:
                raise ResourceNotFoundError(_error_messages.RESOURCE_NOT_FOUND_INVESTMENT_DATA_LIST_ID)

        elif self.search_criteria is not None and len(self.search_criteria.strip()) > 0:
            id_list = self._get_investment_ids_by_search_id()
            if not id_list:
                raise ResourceNotFoundError(_error_messages.RESOURCE_NOT_FOUND_INVESTMENT_DATA_SEARCH_CRITERIA_ID)

        elif self.search_criteria_object:
            id_list = self._get_investments_by_search_criteria_object()
            if not id_list:
                raise ResourceNotFoundError(_error_messages.RESOURCE_NOT_FOUND_INVESTMENT_DATA_SEARCH_CRITERIA_CONDITION)

        return id_list

    def _get_investment_ids_by_list_id(self) -> list:
        investment_list_dict = self._get_investment_list()
        investment_id_list = investment_list_dict.get("idList", []) if investment_list_dict is not None else []
        investment_ids = []
        if investment_id_list:
            investment_ids = list(map(lambda x: self._concat_id_and_type(x), investment_id_list))
        return investment_ids

    def _get_investment_ids_by_search_id(self) -> list:
        investment_list = self._get_investments_by_search_id()
        investment_ids = []
        if investment_list:
            mapper: Callable[[Dict[str, Any]], Any] = lambda x: x.get("id", "")
            investment_ids = list(map(mapper, investment_list))
        return investment_ids

    def _get_investment_list(self) -> Dict[str, Any]:
        # sourcetype should be "Desktop" to retrieve secids for list items
        # The alternative is "Cloud", which retrieves performanceids.
        # We need secids for joining with Datalake tables
        url = f"{_config.object_service_url()}objectapi/v1/lists/{self.list_id}?sourcetype=Desktop"
        response_json: Dict[str, Any] = _direct_api_request("GET", url)

        investment_list: List[Dict[str, Any]] = response_json.get("lists", [])
        if len(investment_list) > 0:
            return investment_list[0]
        return dict()

    def _get_investments_by_search_id(self) -> list:
        url = f"{_config.searches_service_url()}v1/searches/{self.search_criteria}"
        response_json = _direct_api_request("GET", url)
        search_criteria = response_json.get("content", dict())
        if len(search_criteria) > 0:
            search_result: list = execute_search(search_criteria)
            return search_result
        return []

    def _get_investments_by_search_criteria_object(self) -> list:
        if self.search_criteria_object is not None and len(self.search_criteria_object) > 0:
            if "universeId" not in self.search_criteria_object.keys() or self.search_criteria_object["universeId"] == "":
                raise BadRequestException("'universeId' is a required property in the search criteria object.")
            if "criteria" not in self.search_criteria_object.keys() or self.search_criteria_object["criteria"] == []:
                raise BadRequestException("'criteria' is a required property in the search criteria object.")
            investment_data = execute_search(self.search_criteria_object)
            return [investmentId["id"] for investmentId in investment_data if "id" in investmentId.keys()]
        return []

    def _concat_id_and_type(self, investment: dict) -> str:
        investment_id = investment.get("id", "").strip()
        investment_type = investment.get("3xType", None)
        investment_type = investment_type.strip() if investment_type else ""
        if investment_type and investment_type.upper() != "CA":
            return f"{investment_id};{investment_type}"
        return f"{investment_id}"


    def generate_investment_source(self) -> dict:
        investment_source = {}
        if self.category_id is not None and len(self.category_id.strip()) > 0:
            if ";" not in self.category_id:
                raise BadRequestException(BAD_REQUEST_ERROR_INVALID_CATEGORY_ID)
            investment_source["type"] = "MORNINGSTAR_CATEGORY"
            investment_source["id"] = self.category_id

        elif self.investment_ids:
            investment_source["type"] = "SECIDS"
            investment_source["id"] = ",".join(self.investment_ids)

        elif self.list_id is not None and len(self.list_id.strip()) > 0:
            investment_source["type"] = "LIST"
            investment_source["id"] = self.list_id

        elif self.search_criteria is not None and len(self.search_criteria.strip()) > 0:
            investment_source["type"] = "SEARCH"
            investment_source["id"] = self.search_criteria

        if not investment_source:
            raise BadRequestException(BAD_REQUEST_ERROR_NO_INVESTMENT_SOURCE)

        return investment_source



import os
import simplejson as json
import functools
from pandas import DataFrame
from typing import Optional, List, Dict, Any, Callable

from ...direct._api import _direct_api_request
from .._config import _Config
from .._exceptions import BadRequestException, ApiResponseException
from .._data_objects import Investments, DataPoints
from .._config_key import ALL_ASSET_FLOW_DATA_POINTS

from ._common import _get_investment_ids, _get_data_points
from ._data import InvestmentDataRequest, InvestmentDataResults, Column

_config = _Config()


class NormalDataProvider:
    @staticmethod
    def build_request(investment_object: Investments, data_point_object: DataPoints, display_name: bool = False) -> InvestmentDataRequest:
        investment_id_list = _get_investment_ids(investment_object)
        data_points = _get_data_points(investment_object, data_point_object, display_name)
        normal_data_points = _filter_out_data_points_by_id(data_points, filter_out_list=ALL_ASSET_FLOW_DATA_POINTS)
        return InvestmentDataRequest(investment_id_list, normal_data_points)

    @staticmethod
    def run_request(req: InvestmentDataRequest) -> InvestmentDataResults:
        raw_data = _get_normal_data(req.investment_ids, req.data_points)
        return _parse_raw_normal_data_values(raw_data)


def _filter_out_data_points_by_id(df: DataFrame, filter_out_list: list) -> DataFrame:
    return df[~df["datapointId"].isin(filter_out_list)].reset_index().drop(["index"], axis=1)


def _get_normal_data(investment_ids: list, normal_data_point_settings: DataFrame) -> dict:
    normal_resp: dict = dict()
    if normal_data_point_settings is not None and not normal_data_point_settings.empty:
        if not investment_ids:
            raise BadRequestException("No investments.")
        data_point_list = normal_data_point_settings.to_dict(orient="records")
        postbody = {
            "datapoints": data_point_list,
            "investments": list(map(lambda x: {"id": x}, investment_ids)),
        }
        resp = _request_investment_data(postbody)
        if resp and isinstance(resp, dict):
            normal_resp.update(resp)

        updated_data_point_list = []

        for data_point, normal_data_point in zip(data_point_list, normal_resp["datapoints"]):
            alias_exists = data_point["alias"] == normal_data_point["alias"]
            use_display_name = "displayName" in normal_data_point_settings.columns and data_point["displayName"]

            if use_display_name:
                name = data_point["displayName"]
            else:
                name = normal_data_point["name"]

            data_point_entry = {**normal_data_point, "name": name}
            updated_data_point_list.append(data_point_entry)

        normal_resp["datapoints"] = updated_data_point_list

    return normal_resp


def _request_investment_data(params: dict) -> Dict[str, Any]:
    if os.getenv("FEED_ID"):
        url = _config.investment_data_service_url()
    else:
        url = f"{_config.securitydata_service_url()}v1/data?includeAdditionalDps=false"

    response_json: Dict[str, Any] = _direct_api_request("POST", url, data=json.dumps(params, ignore_nan=True))
    return response_json


def _parse_raw_normal_data_values(response_json: dict) -> InvestmentDataResults:
    investment_results = InvestmentDataResults()
    data_points = response_json.get("datapoints", [])
    for inv in response_json.get("investments", []):
        investment_id = inv.get("id", "")
        if _with_entitled(inv) is False:
            investment_results.add_meta_data(investment_id, "entitled", False)
            continue
        value_map = dict(list(map(lambda x: (x.get("alias", ""), x), inv.get("values", []))))
        for dp in data_points:
            alias = dp.get("alias", "")
            value = value_map.get(alias, None)
            investment_results.add_meta_data(investment_id, alias, dp)
            add_column_data = functools.partial(investment_results.add_column_data, investment_id, alias)

            if _with_entitled(value) is False:
                add_column_data(_convert_data_point_to_column(dp, None))
                continue

            if dp.get("isTsdp", False):
                add_column_data(_convert_ts_data_point_to_columns(dp, value))
                continue

            if "MULTIPLE" == dp.get("nonstandardDisplayType", ""):
                add_column_data(_convert_multiple_display_type_data_point_to_column(dp, value))
            else:  # current data_point
                add_column_data(_convert_data_point_to_column(dp, value))
    return investment_results


def _convert_data_point_to_column(data_point: dict, value: Optional[dict]) -> list:
    # { "alias": "HS05A_4", "value": "3", "text": "Large Growth" }
    data_point_name = data_point.get("name", "")
    if value is None or not isinstance(value, dict):
        return [Column(name=data_point_name, value=value)]
    if _is_mstar_ip_data_point(data_point):
        return [Column(name=data_point_name, value=value.get("value")), Column(name=_get_mstar_ip_data_point_text_column_name(data_point), value=value.get("text"))]
    else:
        return [Column(name=data_point_name, value=value.get("text", value.get("value")))]


def _convert_ts_data_point_to_columns(data_point: dict, value: dict) -> list:
    # { "alias": "HS05A_3", "value": [ { "index": 0, "text": "Large Growth", "value": "3" }, { "index": 1, "text": "Large Growth", "value": "3" } ] }
    column_data: list = list()
    value_map = dict(list(map(lambda x: (x.get("index", ""), x), value.get("value", [])))) if isinstance(value, dict) else dict()
    for i, tp in enumerate(data_point.get("timePeriods", [])):
        value_dict: dict = value_map.get(i, dict())
        if _with_entitled(value_dict) is False:
            value_dict = dict()
        if _is_mstar_ip_data_point(data_point):
            column_data.append(Column(name=_get_ts_column_name(data_point, _get_date(tp)), value=value_dict.get("value")))
            column_data.append(Column(name=_get_mstar_ip_data_point_text_column_name(data_point, _get_date(tp)), value=value_dict.get("text")))
        else:
            column_data.append(Column(name=_get_ts_column_name(data_point, _get_date(tp)), value=value_dict.get("text", value_dict.get("value"))))
    return column_data


def _convert_multiple_display_type_data_point_to_column(data_point: dict, value: dict) -> list:
    #  { "alias": "HS05A_3", "value": [{ "value": "[1996-11-01 -- ] William C. Nygren" },{ "value": "[2000-03-21 -- 2012-08-01] Henry Berghoef" }] }
    data_point_name = data_point.get("name", "")
    val: Optional[List[Dict[str, Any]]] = value.get("value") if isinstance(value, dict) else value
    if isinstance(val, list):
        get_value: Callable[[Dict[str, Any]], Any] = lambda x: x["value"]
        v: List[Any] = list(map(get_value, val))
        values: Optional[List[Any]] = v if len(v) > 0 else None
        return [Column(name=data_point_name, value=values)]
    else:
        return [Column(name=data_point_name, value=val)]


def _is_mstar_ip_data_point(data_point: dict) -> bool:
    mstar_ip_type = data_point.get("mstarIpType", "") if data_point is not None else ""
    return mstar_ip_type is not None and len(mstar_ip_type) > 0


def _get_date(timeperiod: dict) -> str:
    start_date = timeperiod.get("startDate", "")
    end_date = timeperiod.get("endDate", "")
    return f"{start_date} to {end_date}" if len(start_date) > 0 else f"{end_date}"


def _get_ts_column_name(data_point: dict, date: str) -> str:
    data_point_name = _get_display_name(data_point)
    return f"{data_point_name} {date}"


def _get_mstar_ip_data_point_text_column_name(data_point: dict, date: Optional[str] = None) -> Any:
    resp = _get_display_name(data_point) + " - display text"
    if date is not None and len(date) > 0:
        resp = resp + " " + date
    return resp


def _get_display_name(data_point: dict) -> str:
    if data_point.get("displayName") is not None:
        return str(data_point.get("displayName"))
    return data_point.get("name", "")


def _with_entitled(data: dict) -> bool:
    # NBI-523: Rollback NBI-433
    # if data:
    #     entitled = data.get("entitled", True)
    #     is_allowed_to_be_exported = data.get("isAllowedToBeExported", True)
    #     if entitled is False or is_allowed_to_be_exported is False:
    #         return False
    return True



from datetime import datetime
import uuid
from ._file import File

from pandas import DataFrame
from io import StringIO

from typing import Any, Optional
import boto3
from botocore.exceptions import ClientError


class CSVFile(File):
    def __init__(self, df: DataFrame) -> None:
        self.csv_buffer = StringIO()
        super().__init__(df)
        self.df.to_csv(self.csv_buffer, index=False, header=False)

    def upload_to_s3(self, bucket: Optional[str]) -> Any:
        s3_resource = boto3.resource("s3")
        path = self.create_s3_file_path()
        try:
            s3_resource.Object(bucket, path).put(Body=self.csv_buffer.getvalue())
        except ClientError as e:
            raise e
        return path

    def create_s3_file_path(self) -> Any:
        basename = "temp_csv"
        obj = uuid.uuid4()
        suffix = datetime.now().strftime("%y%m%d_%H%M%S") + ".csv"
        filename = "_".join([basename, suffix])
        path = f"{obj}/{filename}"
        return path



import simplejson as json
from dataclasses import dataclass
from pandas import DataFrame

from typing import Optional, List, Any
from .._exceptions import BadRequestException
from .._config import _Config
from ..data_type import PeerGroupMethodology, Order
from .._data_objects import DataPoints, Investments
from ..._base import _logger
from .._exceptions import ResourceNotFoundError, InternalServerError
from .._base_api import APIBackend
from .._error_messages import BAD_REQUEST_ERROR_INVALID_PERCENTILES

_config = _Config()


class PeerGroupAPIBackend(APIBackend):
    """
    Subclass to call the Peer Group Data API and handle any HTTP errors that occur.
    """

    def __init__(self) -> None:
        super().__init__()

    def _handle_custom_http_errors(self, res: Any) -> Any:
        """
        Handle HTTP errors with custom error messages
        """
        response_message = res.json()["message"]
        if res.status_code == 404:
            _logger.debug(f"Resource Not Found Error: {res.status_code} {response_message}")
            raise ResourceNotFoundError from None

        if res.status_code == 500:
            _logger.debug(f"Something went wrong: {res.status_code} {response_message}")
            raise InternalServerError(response_message) from None


_peer_group_api_request = PeerGroupAPIBackend()


@dataclass(frozen=True)
class PeerGroupRequest:
    investment_source: dict
    data_points: DataFrame
    order: Order
    percentiles: Optional[List[int]] = None
    methodology: Optional[PeerGroupMethodology] = None


@dataclass(frozen=True)
class PeerGroupResponse:
    req: PeerGroupRequest
    breakpoints: list


class PeerGroupProvider:
    def build_request(
        self,
        investment_object: Investments,
        data_point_object: DataPoints,
        order: Order,
        percentiles: Optional[List[int]] = None,
        methodology: Optional[PeerGroupMethodology] = None,
    ) -> PeerGroupRequest:
        # do not support dataset id
        data_point_object.data_set_id = None
        methodology = _get_methodology_value(methodology)
        _validate_percentiles(percentiles)

        investment_source = investment_object.generate_investment_source()
        data_point_data_frame = _get_data_points(data_point_object)

        return PeerGroupRequest(investment_source, data_point_data_frame, order, percentiles, methodology)

    def run_request(self, req: PeerGroupRequest) -> PeerGroupResponse:
        breakpoints = _get_breakpoints(
            investment_source=req.investment_source, data_point_data_frame=req.data_points, order=req.order.value, methodology=req.methodology, percentiles=req.percentiles
        )
        return PeerGroupResponse(req, breakpoints)

    def build_data_frame(self, resp: PeerGroupResponse) -> DataFrame:
        return _convert_to_breakpoints_data_frame(resp.breakpoints)


def _get_breakpoints(investment_source: dict, data_point_data_frame: DataFrame, order: Any, methodology: Optional[str] = None, percentiles: Optional[list] = None) -> list:
    breakpoints = []
    postbody = {"investmentSource": investment_source, "datapoints": data_point_data_frame.to_dict(orient="records"), "percentiles": percentiles, "order": order}
    url = f"{_config.peergroup_service_url()}peergroupservice/v1/breakpoints"
    if methodology is not None and len(methodology.strip()) > 0:
        url = f"{url}?methodology={methodology}"
    resp = _peer_group_api_request.do_post_request(url, json.dumps(postbody, ignore_nan=True))
    if resp and isinstance(resp, dict):
        breakpoints = resp.get("breakpoints", [])
    return breakpoints


def _convert_to_breakpoints_data_frame(breakpoints: list) -> DataFrame:
    breakpoint_list = []
    for breakpoint in breakpoints:
        breakpoint_dict = dict()
        breakpoint_dict["Alias"] = breakpoint.get("alias", None)
        time_period = breakpoint.get("timePeriod", dict())
        breakpoint_dict["StartDate"] = time_period.get("startDate", dict())
        breakpoint_dict["EndDate"] = time_period.get("endDate", dict())
        values = breakpoint.get("values", list())
        for value in values:
            breakpoint_dict[value.get("percentile", None)] = value.get("value", None)
        breakpoint_list.append(breakpoint_dict)
    return DataFrame(breakpoint_list)


def _get_methodology_value(methodology: Optional[PeerGroupMethodology] = None) -> Any:
    if methodology is not None and isinstance(methodology, PeerGroupMethodology):
        return methodology.value
    return None


def _validate_percentiles(percentiles: Optional[list] = None) -> None:
    if not percentiles:
        return

    for per in percentiles:
        if per is None or not isinstance(per, int) or per < 1 or per > 100:
            raise BadRequestException(BAD_REQUEST_ERROR_INVALID_PERCENTILES)


def _get_data_points(data_point_object: DataPoints) -> DataFrame:
    data_point_object._validate_data_point_ids_and_alias()
    return data_point_object.get_peer_group_data_points()



import simplejson as json

from pandas import DataFrame
from typing import List, Any, Optional, Union, Dict

from . import _decorator, _utils, _error_messages
from .._base import _logger
from ._config import _Config
from ._data_objects import Investments
from ._data_point import _request_asset_flow_data_points
from ._exceptions import BadRequestException, QueryLimitException, ResourceNotFoundError
from ._base_api import APIBackend
from .._utils import _get_user_cells_quota, _get_data_points_total_columns
from ._data_type import DryRunResults


_config = _Config()


class AssetFlowAPIBackend(APIBackend):
    """
    Subclass to call the Asset Flow Data API and handle any HTTP errors that occur.
    """

    def __init__(self) -> None:
        super().__init__()

    def _handle_custom_http_errors(self, res: Any) -> Any:
        """
        Handle HTTP errors with custom error messages
        """
        response_message = res.json()["message"]
        if res.status_code == 403 and "Exceed query limitation." in response_message:
            _logger.debug(f"Query Limit Exception: {res.status_code} {response_message}")
            query_limit = self._get_security_data_query_limit()
            raise QueryLimitException(query_limit) from None
        elif res.status_code == 404:
            _logger.debug(f"Resource Not Found Error: {res.status_code} {response_message}")
            raise ResourceNotFoundError(_error_messages.RESOURCE_NOT_FOUND_ASSET_FLOW) from None

    def _get_security_data_query_limit(self) -> str:
        try:
            url = f"{_config.securitydata_service_url()}v1/limitation/summary"
            res = self.do_get_request(url)
            return str(res["limitationTotal"])
        except Exception as e:
            _logger.error(f"Error getting security data query limit: {e}")
            raise QueryLimitException from None


_asset_flow_api_request = AssetFlowAPIBackend()


@_decorator.typechecked
def get_asset_flow_markets() -> DataFrame:
    """Returns all investment markets that can be used to retrieve asset flow data. For example,
    "US Open-end & ETFs ex MM ex FoF".

    Returns:
        DataFrame: A DataFrame object with asset flow markets data. DataFrame columns include:

        * marketId
        * marketName
        * currency

    :Examples:

    ::

        import morningstar_data as md

        df = md.direct.get_asset_flow_markets()
        df

    :Output:
        ========  ===============================  ========
        marketId  marketName                       currency
        ========  ===============================  ========
        5         US Open-end & ETF ex MM ex FoF   USD
        6         US Open-end, ETF, and MM ex FoF  USD
        ...
        ========  ===============================  ========

    Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    url = f"{_config.securitydata_service_url()}v1/assetflow/all-markets"
    response_json = _asset_flow_api_request.do_get_request(url)
    if response_json and isinstance(response_json, list):
        return DataFrame(response_json)[["marketId", "marketName", "currency"]]
    else:
        return DataFrame({"marketId": [], "marketName": [], "currency": []})


def get_asset_flow_data_points() -> DataFrame:
    """Returns all available data points related to asset flows.

    Returns:
        DataFrame: A DataFrame object with asset flow data points data. DataFrame columns include:

        * datapointId
        * datapointName
        * asOfDate
        * alias
        * startDate
        * endDate
        * frequency

    Examples:

    ::

        import morningstar_data as md

        df = md.direct.get_asset_flow_data_points()
        df

    :Output:
        ===========  =============  ==========  ==========================================  =========  =======  =========
        datapointId  datapointName  asOfDate    alias                                       startDate  endDate  frequency
        ===========  =============  ==========  ==========================================  =========  =======  =========
        TNA0M        XXX            2021-09-30  Total Net Assets-Market Value(Share Class)  None       None     None
        ...
        ===========  =============  ==========  ==========================================  =========  =======  =========

    """
    response_json = _request_asset_flow_data_points()
    if response_json and isinstance(response_json, list):
        for settings in response_json:
            settings["datapointName"] = settings.pop("name")
            settings["alias"] = settings["datapointName"]

        settings = _utils._extract_data(response_json)
        settings_data_frame = DataFrame(settings)

        column_list = settings_data_frame.columns.tolist()
        if ("datapointId" in column_list) & (column_list.index("datapointId") != 0):
            data_point_id_col = settings_data_frame["datapointId"]
            settings_data_frame = settings_data_frame.drop("datapointId", axis=1)
            settings_data_frame.insert(0, "datapointId", data_point_id_col)

        column_list = settings_data_frame.columns.tolist()
        if ("datapointName" in column_list) & (column_list.index("datapointName") != 1):
            data_point_name_col = settings_data_frame["datapointName"]
            settings_data_frame = settings_data_frame.drop("datapointName", axis=1)
            settings_data_frame.insert(1, "datapointName", data_point_name_col)

        settings_data_frame = settings_data_frame.where(settings_data_frame.notnull(), None)
        return settings_data_frame
    else:
        return DataFrame()


@_decorator.typechecked
def get_asset_flow(
    market_id: str,
    data_point_settings: DataFrame,
    investments: Optional[Union[List[str], str, Dict[str, Any]]] = None,
    dry_run: Optional[bool] = False,
) -> Union[DataFrame, DryRunResults]:
    """Get asset flow data for a market of investments or specific investments within a market.

    Args:
        market_id(:obj:`str`): A numeric code representing a broad market of investments. For example, the code for "US Open-end & ETF ex MM ex FoF" is "5". Use the `get_asset_flow_markets <./assetflow.html#morningstar_data.direct.get_asset_flow_markets>`_ function to retrieve a full list of codes.

        data_point_settings(:obj:`DataFrame`): A DataFrame of data points with defined settings. Each row represents a data point. Each column is a configurable setting. This DataFrame can be obtained by `retrieving asset flow data points <./assetflow.html#morningstar_data.direct.get_asset_flow_data_points>`_, or by `retrieving data point settings <./lookup.html#morningstar_data.direct.get_data_point_settings>`_.

        investments(:obj:`Union`, `required`): Defines the investments to fetch. Input can be:

            * Investment IDs (:obj:`list`, `optional`): Investment identifiers, in the format of SecId;Universe or just SecId. E.g., ["F00000YOOK;FO","FOUSA00CFV;FO"] or ["F00000YOOK","FOUSA00CFV"]. Use the `investments <./lookup.html#morningstar_data.direct.investments>`_ function to discover identifiers.
            * Investment List ID (:obj:`str`, `optional`): Saved investment list in Morningstar Direct. Use the `get_investment_lists <./lists.html#morningstar_data.direct.user_items.get_investment_lists>`_ function to discover saved lists.
            * Search Criteria  ID (:obj:`str`, `optional`): Saved search criteria in Morningstar Direct. Use the `get_search_criteria <./search_criteria.html#morningstar_data.direct.user_items.get_search_criteria>`_ function to discover saved search criteria.
            * Search Criteria Condition (:obj:`dict`, `optional`): Search criteria definition. See details in the Reference section of `get_investment_data <./investment.html#morningstar_data.direct.get_investment_data>`_ or use the `get_search_criteria_conditions <./search_criteria.html#morningstar_data.direct.user_items.get_search_criteria_conditions>`_ function to discover the definition of a saved search criteria.

        dry_run(:obj:`bool`, `optional`): When True, the query will not be executed. Instead, a DryRunResults object will be returned with details about the query's impact on daily cell limit usage.

    :Returns:

        There are two return types:

        * DataFrame: A DataFrame object with asset flow data. DataFrame columns include ``investmentId`` and data point names, as
        provided in ``data_point_settings``.

        * DryRunResults: Returned if dry_run=True is passed

          * estimated_cells_used: Number of cells by this query
          * daily_cells_remaining_before: How many cells are remaining in your daily cell limit before running this query
          * daily_cells_remaining_after: How many cells would be remaining in your daily cell limit after running this query
          * daily_cell_limit: Your total daily cell limit

    :Examples:

    ::

        import morningstar_data as md
        import pandas

        ASSET_FLOW_DATA_POINT_SETTINGS = [
            {
                "datapointId": "TNA0M",
                "datapointName": "Total Net Assets-Market Value(Share Class)",
                "asOfDate": "2021-08-30",
                "alias": "Total Net Assets-Market Value(Share Class)",
                "startDate": None,
                "endDate": None,
                "frequency": None,
            }
        ]
        settings = pandas.DataFrame(ASSET_FLOW_DATA_POINT_SETTINGS)
        df = md.direct.get_asset_flow(
            investments=["F000010HRO"], market_id="165", data_point_settings=settings
        )
        df

    :Output:
        ============  =======================================================
        investmentId  Total Net Assets-Market Value(Share Class) - 2021-06-30
        ============  =======================================================
        F000010HRO    0.00188
        ============  =======================================================

    Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    if data_point_settings.empty:
        raise BadRequestException("data_point_settings is required.")
    investment_id_list: List[Any] = []

    if investments is not None:
        investment_object = Investments(investments)
        investment_id_list = investment_object.get_investment_ids()

    data_point_settings = data_point_settings.where(data_point_settings.notnull(), None)
    data_point_list = data_point_settings.to_dict(orient="records")

    if dry_run:
        estimated_cells_used = _get_data_points_total_columns(data_point_list) * len(investment_id_list)
        user_cells_quota = _get_user_cells_quota()
        dry_run_results = DryRunResults(
            estimated_cells_used=estimated_cells_used, daily_cells_remaining_before=user_cells_quota["daily_cell_remaining"], daily_cell_limit=user_cells_quota["daily_cell_limit"]
        )
        return dry_run_results

    postbody = {"marketId": market_id, "datapoints": data_point_list}
    if investment_id_list is not None:
        postbody["investments"] = list(map(lambda x: {"id": _remove_univ(x)}, investment_id_list))
    return _get_data(postbody)


def _get_data(data: dict) -> DataFrame:
    try:
        url = f"{_config.securitydata_service_url()}v1/assetflow/data"
        response_json = _asset_flow_api_request.do_post_request(url, json.dumps(data, ignore_nan=True))
        if not response_json or not isinstance(response_json, list):
            return DataFrame()

        data_point_columns: dict = {x.get("alias", "").strip(): [] for x in data.get("datapoints", [])}
        investment_data_list = []
        for investment in response_json:
            data_point_values = dict()
            data_point_values["investmentId"] = investment.get("id", "")
            values = investment.get("values", [])
            for data_point_value in values:
                alias = data_point_value.get("alias", "").strip()
                value_list = data_point_value.get("value", [])
                data_point_column = data_point_columns.get(alias)
                if data_point_column is None:
                    data_point_column = []
                    data_point_columns[alias] = data_point_column

                data_point_column.extend([_concat_str(alias, x.get("date", "")) for x in value_list if _concat_str(alias, x.get("date", "")) not in data_point_column])
                data_point_values.update({_concat_str(alias, x.get("date", "")): x.get("value", None) for x in value_list})
            investment_data_list.append(data_point_values)

        result = DataFrame(investment_data_list)
        column_order = ["investmentId"]
        return _column_reorder(data_point_columns, column_order, result)
    except Exception as e:
        raise e


def _column_reorder(data_point_columns: dict, column_order: list, result: DataFrame) -> DataFrame:
    for alias, column in data_point_columns.items():
        column = sorted(set(column))
        column_order.extend(column)
    return result[column_order]


def _remove_univ(investment_id: Optional[str]) -> Optional[str]:
    sec_id = investment_id
    if sec_id is not None and ";" in sec_id:
        return sec_id.split(";")[0]
    return sec_id


def _concat_str(s1: str, s2: str) -> str:
    return f"{s1} - {s2}"



import dataclasses
import simplejson as json
import uuid
from enum import Enum
from typing import Any, Dict, Optional, List
from ._common import PortfolioDataApiBackend
from .._config import _Config
from ..._base import _logger

_portfolio_api = PortfolioDataApiBackend()
_config = _Config()


class PerformanceSource(Enum):
    Calculated_Based_On_Underlying_Positions = 1
    Imported_Account_Level_Performance = 2


class RebalanceOnType(Enum):
    Calendar_Period_End = 0
    Rolling_Period_Based_On_Portfolio_Date = 1


class RebalanceFrequency(Enum):
    Monthly = 2
    Quarterly = 3
    Semi_Annually = 4
    Annually = 5
    Buy_And_Hold = 7
    Daily = 8


class CalculationType(Enum):
    Earliest_Common = 1
    First_Portfolio_Date = 2
    Earliest_Available = 3


class PerformanceSeriesFrequency(Enum):
    Monthly = 4
    Quarterly = 5
    Daily = 6


class ManagementFeeEffectivePeriodType(Enum):
    Entire_Period = 0
    Custom_Period = 1


class ManagementFeeFrequency(Enum):
    Daily = 1
    Monthly = 2
    Quarterly = 3
    Semi_Annually = 4
    Annually = 5


class BenchmarkType(Enum):
    Morningstar_Security = 1
    User_Defined_Security = 2


class MissingTNAHandling(Enum):
    Roll_Previous_Value_Forward = 1
    Infer_Forward_In_Time = 2
    Infer_Backward_In_Time = 3


class Benchmark:
    benchmark_id: str
    benchmark_type_id: int = 1
    investment_id: str

    def __init__(self, investment_id: str, benchmark_id: str) -> None:
        self.investment_id = investment_id
        self.benchmark_type_id = 1
        self.benchmark_id = benchmark_id


class PortfolioSetting:
    notes: None
    benchmark_id: str
    missing_tna_handling_id: MissingTNAHandling = MissingTNAHandling.Infer_Forward_In_Time
    management_fee_effective_period_type: ManagementFeeEffectivePeriodType = ManagementFeeEffectivePeriodType.Entire_Period
    management_fee_frequency: ManagementFeeFrequency = ManagementFeeFrequency.Annually
    management_fee_for_performance: Optional[float] = None
    risk_free_proxy_id: str = "XIUSA000OC"
    rebalance_frequency_id: RebalanceFrequency = RebalanceFrequency.Buy_And_Hold
    start_date_calculation_type_id: CalculationType = CalculationType.First_Portfolio_Date
    rebalance_based_on: RebalanceOnType = RebalanceOnType.Calendar_Period_End
    performance_source_id: PerformanceSource = PerformanceSource.Calculated_Based_On_Underlying_Positions
    performance_series_frequency_id: PerformanceSeriesFrequency = PerformanceSeriesFrequency.Quarterly
    portfolio_to_id: None
    model_portfolio_id: None
    model_portfolio_name: None
    use_gross_return: bool = False
    real_sma_flag: bool = False
    is_combined_series: bool = False
    combined_series: None
    base_currency: str = "USD"
    secondary_benchmark_id: str = "XIUSA04FMS"
    portfolio_type: Optional[str]
    portfolio_id: Optional[str]
    name: Optional[str]
    benchmarks: List[Benchmark]

    def __init__(
        self,
        name: Optional[str] = None,
        portfolio_type: Optional[str] = None,
        portfolio_id: Optional[str] = None,
    ) -> None:
        """
        Build portfolio settings with default value.
        """
        self.notes = None
        self.benchmark_id = "XIUSA04G92"
        self.missing_tna_handling_id = MissingTNAHandling.Infer_Forward_In_Time
        self.management_fee_effective_period_type = ManagementFeeEffectivePeriodType.Entire_Period
        self.management_fee_frequency = ManagementFeeFrequency.Annually
        self.management_fee_for_performance = None
        self.risk_free_proxy_id = "XIUSA000OC"
        self.rebalance_frequency_id = RebalanceFrequency.Buy_And_Hold
        self.start_date_calculation_type_id = CalculationType.First_Portfolio_Date
        self.rebalance_based_on = RebalanceOnType.Calendar_Period_End
        self.performance_source_id = PerformanceSource.Calculated_Based_On_Underlying_Positions
        self.performance_series_frequency_id = PerformanceSeriesFrequency.Quarterly
        self.portfolio_to_id = None
        self.model_portfolio_id = None
        self.model_portfolio_name = None
        self.use_gross_return = False
        self.real_sma_flag = False
        self.is_combined_series = False
        self.combined_series = None
        self.base_currency = "USD"
        self.secondary_benchmark_id = "XIUSA04FMS"
        self.portfolio_type = portfolio_type
        self.portfolio_id = portfolio_id
        self.name = name

    def get_formatted_json_body(self) -> Dict[Any, Any]:
        """
        Build portfolio settings post body.
        """
        # Generate portfolio benchmark objects request dict.
        _logger.info("Generate portfolio benchmark objects request.")
        benchmark_uuid = str(uuid.uuid4())
        secondary_benchmark_uuid = str(uuid.uuid4())
        risk_free_proxy_uuid = str(uuid.uuid4())

        # Each benchmarks need a UUID.
        self.benchmarks = [
            Benchmark(self.benchmark_id, benchmark_uuid),
            Benchmark(self.secondary_benchmark_id, secondary_benchmark_uuid),
            Benchmark(self.risk_free_proxy_id, risk_free_proxy_uuid),
        ]

        benchmarks = []
        # Generate benchmark dict list.
        for benchmark in self.benchmarks:
            benchmarks.append({"securityId": benchmark.investment_id, "benchmarkTypeId": benchmark.benchmark_type_id, "benchmarkId": benchmark.benchmark_id})
        # Generate portfolio settings with default values .
        _logger.info("Generate portfolio settings with default values")
        to_return = {
            "settings": {
                "portfolio": {
                    "portfolioType": self.portfolio_type,
                    "notes": self.notes,
                    "portfolioId": self.portfolio_id,
                    "benchmarkId": benchmark_uuid,
                    "baseCurrency": self.base_currency,
                    "missingTNAHandlingId": self.missing_tna_handling_id.value,
                    "managementFeeEffectivePeriodType": self.management_fee_effective_period_type.value,
                    "managementFeeFrequency": self.management_fee_frequency.value,
                    "managementFeeForPerformance": self.management_fee_for_performance,
                    "riskFreeProxyId": secondary_benchmark_uuid,
                    "rebalanceFrequencyId": self.rebalance_frequency_id.value,
                    "secondaryBenchmarkId": risk_free_proxy_uuid,
                    "startDateCalculationTypeId": self.start_date_calculation_type_id.value,
                    "rebalanceBasedOn": self.rebalance_based_on.value,
                    "performanceSourceId": self.performance_source_id.value,
                    "performanceSeriesFrequencyId": self.performance_series_frequency_id.value,
                    "portfolio2Id": self.portfolio_to_id,
                    "modelPortfolioId": self.model_portfolio_id,
                    "modelPortfolioName": self.model_portfolio_name,
                    "useGrossReturn": self.use_gross_return,
                    "realSMAFlag": self.real_sma_flag,
                    "isCombinedSeries": self.is_combined_series,
                    "combinedSeries": self.combined_series,
                    "name": self.name,
                },
                "benchmarks": benchmarks,
            }
        }
        _logger.debug(f"Generate portfolio settings with default values. {to_return}")
        return to_return

    def save_portfolio_settings(self) -> Any:

        url = f"{_config.portfolio_service_url()}portfoliodataservice/v1/portfolios/update-portfolio-settings"
        _logger.debug(f"Save portfolio settings with portfolio API: {url}")
        response_json = _portfolio_api.do_post_request(url, json.dumps(self.get_formatted_json_body(), ignore_nan=True))
        _logger.debug(f"Save portfolio API response: {response_json}")
        return response_json



import os
from math import isnan
import requests
import polling2
from ._config import _Config
from typing import Any, Dict, Optional, Union
from ._exceptions import ApiRequestException, ApiResponseException, InternalServerError
from ..direct._error_messages import INTERNAL_SERVER_ERROR
from .._base import _logger
from tenacity import retry, stop_after_attempt, wait_exponential
from ._utils import get_bytes
from .._version import __version__
import uuid


_config = _Config()

POLL_STEP = 1  # seconds
POLL_TIMEOUT = 900  # seconds (15 minutes)
HEADER_STATUS = "proxy_response_status"


def get_poll_step_from_headers(headers: Dict[str, Any]) -> int:
    return int(headers["POLL_STEP"]) if "POLL_STEP" in headers else POLL_STEP


def get_poll_timeout_from_headers(headers: Dict[str, Any]) -> int:
    return int(headers["POLL_TIMEOUT"]) if "POLL_TIMEOUT" in headers else POLL_TIMEOUT


def is_task_complete(response: Any) -> bool:
    if "proxy_response_status" not in response.headers:
        return False
    if response.headers["proxy_response_status"].upper() == "SUCCESS" or response.headers["proxy_response_status"].upper() == "FAILURE":
        return True
    else:
        return False


@retry(reraise=True, stop=stop_after_attempt(8), wait=wait_exponential(multiplier=2, min=1, max=120))  # type: ignore
def request_with_retry(method: str, url: str, headers: Dict[str, Any], verify: bool, data: Optional[Union[Dict[Any, Any], str]] = None) -> requests.Response:

    STATUS_FORCE_LIST = [500, 501, 502, 503, 504]
    MAX_REDIRECTS = 5
    REDIRECT_CODES = [301, 307]
    try:
        res = requests.request(method, url, headers=headers, verify=verify, data=data, allow_redirects=False)

        redirect_count = 0

        while res.status_code in REDIRECT_CODES and redirect_count <= MAX_REDIRECTS:

            redirected_url = res.headers["Location"]
            _logger.info(f"Redirection to {redirected_url} for {method} request to {url}")

            res = requests.request(method, redirected_url, headers=headers, verify=verify, data=data, allow_redirects=False)

            redirect_count += 1

    except Exception as e:
        _logger.error(f"Retry Error: Exception making request. method: {method},url: {url},headers: {headers} , error: {e}")
        if "proxy_request" in url:
            raise ApiRequestException(f"Failed connecting to {url}, error: {e}")
        else:
            raise ApiResponseException(f"Failed connecting to {url}, error: {e}")

    if res.status_code in STATUS_FORCE_LIST:
        # Build a message from the response if we can
        try:
            res_json = res.json()
        except requests.JSONDecodeError:
            raise InternalServerError(f"{INTERNAL_SERVER_ERROR}. Got status code: {res.status_code} for method: {method} to {url}.")

        error_msg = f"{INTERNAL_SERVER_ERROR}. Got status code: {res.status_code} for method: {method} to {url}."

        try:
            hint = res_json["_meta"]["hint"]
            error_msg = error_msg + f" hint: {hint}"
        except KeyError:
            pass

        try:
            message = res_json["message"]
            error_msg = error_msg + f" message: {message}"
        except KeyError:
            pass

        raise InternalServerError(error_msg)

    return res


# This function checks whether the response from a function is a presigned URL, and if it is, it downloads the content from that URL and returns the response object.
# If the response is not a presigned URL, it returns the response object.
def download_if_signed_url(response: requests.Response) -> requests.Response:
    if response.headers.get("download_presigned_url") == "True":
        download_url = response.content
        _logger.info(f"Downloading from s3 at the url {download_url}")
        res_s3_byte_object = get_bytes(download_url)
        # Note that _content is a private attribute, so we should not normally modify it. However, in this case, we need to modify it to change the value of the content attribute.
        response._content = res_s3_byte_object.content
        return response
    else:
        return response


def make_request(method: str, url: str, headers: Dict[str, Any], verify: bool, data: Optional[Union[Dict[Any, Any], str]] = None) -> requests.Response:
    request_id = None
    poll_step = POLL_STEP
    poll_timeout = POLL_TIMEOUT
    try:
        _logger.info("Queueing request")
        _logger.debug(f"Queueing request: {method} {url}")
        headers["md-package-version"] = __version__

        if "X-API-RequestId" not in headers.keys() and "x-api-requestid" not in headers.keys():
            headers["X-API-RequestId"] = str(uuid.uuid4())

        if os.getenv("REQUEST_ID", "") != "":
            headers["X-API-RunId"] = os.getenv("REQUEST_ID")

        if os.getenv("FEED_ID"):
            headers["x-feed-id"] = os.getenv("FEED_ID")     

        request_uuid = headers.get("X-API-RequestId", headers.get("x-api-requestid"))
        _logger.debug(f"Updated Request ID: {request_uuid}")
        res = request_with_retry(method, url, headers=headers, verify=verify, data=data)
        res.raise_for_status()

        _logger.info("Got valid response from queueing request")
        _logger.debug(f"Got {res.status_code} response from queueing request.")

        _logger.debug(f"Response headers: {res.headers}")

        proxy_status = res.headers[HEADER_STATUS].upper()
        if proxy_status == "SUCCESS":
            _logger.debug("Request is already completed. Returning the response.")
            return res
        elif proxy_status == "QUEUED":
            _logger.debug("Request is still queued. Getting the request id")
            request_id = res.json()["id"]
            _logger.debug(f"Client Request ID {request_uuid}. MD-API Task ID {request_id}")
            _logger.info(f"Request id is {request_id}")
            poll_step = get_poll_step_from_headers(res.headers)
            poll_timeout = get_poll_timeout_from_headers(res.headers)

    except requests.ConnectionError as e:
        _logger.error(repr(e))
        raise ApiRequestException(f"Failed connecting to {url}")
    except requests.HTTPError as e:
        # If the request is already completed, return the response.
        # The response is considered successful if the proxy_response_status in the header is "completed"
        # Every response from MD API has a proxy_response_status header
        _logger.info("Got HTTPError from queueing request")
        # Access the value of "proxy_response_status" in res.headers dictionary, with a default value of an empty string ("") if the key is not found
        # if res.headers.get("proxy_response_status", "").upper() == "SUCCESS":
        #     _logger.debug("Request is already completed. Returning the response.")
        #     return res

        _logger.error(repr(e))
        # raise ApiRequestException(f"Failed queueing the request. Got {res.status_code}")
        raise ApiRequestException(f"Failed queueing the request.")

    response_url = f"{_config._MD_API}proxy_response/{request_id}"
    _logger.debug(f"Response URL is {response_url}")

    # Poll the response URL until the request is completed
    # Return the response object as it is as exception handling is done in the calling function
    # Response is considered successful if the proxy_response_status in the header is "completed"

    # time.sleep(5)

    # return requests.get(response_url, headers=headers, verify=verify)

    response = polling2.poll(
        lambda: request_with_retry("GET", response_url, headers=headers, verify=verify),
        step=poll_step,
        timeout=poll_timeout,
        check_success=lambda response: is_task_complete(response),
    )
    return download_if_signed_url(response=response)



import datetime
import time

from uuid import UUID
from pandas import DataFrame
from typing import List, Any, Dict, Callable
from .._base import _logger
from ._exceptions import BadRequestException
from ._error_messages import BAD_REQUEST_ERROR_INVALID_DATE_FORMAT
import requests

from ._config_key import REQUIRED_DATA_POINT_SETTINGS, VISIBLE_DATA_POINT_SETTINGS


FORMAT_DATE = "%Y-%m-%d"


def _reduce_list_data(lists: List[Dict[str, Any]], props_dict: Dict[str, str]) -> List[Dict[str, Any]]:
    filtered_lists, id_for, name = [], props_dict["id"], props_dict["name"]
    for item in lists:
        data: Dict[str, Any] = {"id": item[id_for], "name": item[name]}
        filtered_lists.append(data)

    return filtered_lists


def _extract_data(data: list) -> List[Any]:
    values = []

    for item in data:
        value = {}
        for k, v in item.items():
            if isinstance(v, dict):
                value.update(v)
            else:
                value[k] = v
        values.append(value)

    return values


def _empty_to_none(df: DataFrame) -> DataFrame:
    return df.where((df.notnull()) & (df != ""), None)


def _get_iso_today() -> str:
    d: datetime.date = datetime.date.today()
    format_iso: str = d.isoformat()
    return format_iso


def _mapper_data_frame_return_list(df: DataFrame, get_values: Callable) -> List:
    data = [entry for entry in df.to_dict(orient="index").values()]
    items: List = [get_values(key, value, row) for row in data for key, value in row.items()]
    filter_not_none: Callable = lambda x: x is not None
    return list(filter(filter_not_none, items))


def _mapper_data_frame_return_dict(df: DataFrame, get_values: Callable) -> Dict:
    result: Dict = {}

    def get_values_for_list(column: str, value: Any, row: Dict[str, Any]) -> None:
        get_values(result, column, value, row)
        return None

    _mapper_data_frame_return_list(df, get_values_for_list)

    return result


def _rename_data_frame_column(df: DataFrame, source: str, target: str) -> DataFrame:
    return df.rename(columns={source: target})


def _filter_data_frame_column_by_setting(df: DataFrame) -> DataFrame:
    data_point_settings = REQUIRED_DATA_POINT_SETTINGS + VISIBLE_DATA_POINT_SETTINGS
    return df.filter(items=data_point_settings)


def _reindex_data_frame_column(df: DataFrame, target_column: str, new_index: int) -> DataFrame:
    column_list = df.columns.tolist()

    if (target_column in column_list) and (column_list.index(target_column) != new_index):
        data_point_id_col = df[target_column]
        df.drop(target_column, axis=1, inplace=True)
        df.insert(new_index, target_column, data_point_id_col)

    return df


def _is_uuid(uuid_str: str, version: int = 4) -> bool:
    try:
        uuid_obj = UUID(uuid_str, version=version)
    except ValueError:
        return False
    return str(uuid_obj) == uuid_str


def _format_date(date: str) -> str:
    try:
        return time.strftime(FORMAT_DATE, time.strptime(date, FORMAT_DATE))
    except Exception as e:
        _logger.debug(f"Date format error: {e}")
        raise BadRequestException(BAD_REQUEST_ERROR_INVALID_DATE_FORMAT) from None


def get_bytes(url: str) -> requests.Response:
    """
    Downloads bytes from S3 presigned url
    """
    res = requests.get(url)
    res.raise_for_status()
    return res



import inspect
import uuid
import re
import copy
import pandas as pd
import numpy as np
import functools
from datetime import datetime

from typing import Optional, List, Union, Any, Dict

from pandas import Series

from .._config_key import ALL_ASSET_FLOW_DATA_POINTS
from .._exceptions import BadRequestException
from ...direct.user_items.data_set import (
    _get_user_data_set,
    _get_defaultview_data_points,
    _get_binding_data_set,
    _get_user_data_set_data_points,
)
from .._data_point import (
    _data_point_request_builder,
    _get_data_point_details,
    _get_asset_flow_data_points_by_ids,
)
from .._utils import _is_uuid
from .._error_messages import (
    BAD_REQUEST_ERROR_NO_DATA_POINT_ID_OR_ALIAS,
    BAD_REQUEST_ERROR_ALIAS_DUPLICATED,
    BAD_REQUEST_ERROR_INCLUDE_NON_CALCULATION_DATA_POINTS,
    BAD_REQUEST_ERROR_INCLUDE_DIFF_CALCULATION_DATA_POINTS,
    BAD_REQUEST_ERROR_NO_DATA_POINTS,
)

DatapointIdType = Optional[str]
DatapointIdsType = Optional[List[str]]
DataPointsSettingsType = Optional[pd.DataFrame]
DataPointsType = Union[DatapointIdType, DatapointIdsType, DataPointsSettingsType]

DefaultDataPoints = [{"datapointId": "HP010"}]


class DataPoints:
    def __init__(self, data_points_object: DataPointsType) -> None:
        self.data_set_id: DatapointIdType = None
        self.data_point_ids: DatapointIdsType = None
        self.data_point_settings: DataPointsSettingsType = None
        self._parse_data_points_object(data_points_object)

    def _parse_data_points_object(self, data_points_object: DataPointsType) -> None:
        if isinstance(data_points_object, str):
            self.data_set_id = data_points_object
        elif isinstance(data_points_object, list):
            if self._is_list_of_dicts(data_points_object):
                self.data_point_ids = data_points_object
            else:
                try:
                    self.data_point_settings = self._convert_morningstar_data_points_object(data_points_object)
                except:
                    ValueError("Invalid data points object")
        elif isinstance(data_points_object, pd.DataFrame):
            self.data_point_settings = data_points_object
        elif data_points_object is not None:
            raise ValueError("Invalid data points object")

    def validate_data_point_ids(self) -> None:
        if self.data_point_ids is not None:
            settings_data_frame = pd.DataFrame(self.data_point_ids)
            if settings_data_frame.empty:
                raise BadRequestException("data_point_ids is empty.")
            elif "datapointId" not in settings_data_frame.columns:
                raise BadRequestException("datapointId is required in data_point_ids.")
        elif self.data_point_settings is not None:
            if self.data_point_settings.empty:
                raise BadRequestException("data_point_settings is empty.")
            elif "datapointId" not in self.data_point_settings.columns:
                raise BadRequestException("datapointId is required in data_point_settings.")

    @functools.lru_cache()
    def get_data_points(
        self,
        list_id: Optional[str] = None,
        search_criteria_id: Optional[str] = None,
        display_name: bool = False,
    ) -> pd.DataFrame:
        data_point_list = []
        if self.data_point_ids:
            settings_data_frame = pd.DataFrame(self.data_point_ids)
            data_point_list = self._convert_data_point_data_frame(settings_data_frame)

            data_point_id_set = set(settings_data_frame["datapointId"])
            valid_data_point_id_set = set(map(lambda x: x["datapointId"], data_point_list))
            invalid_data_point_id_set = data_point_id_set - valid_data_point_id_set

            if data_point_id_set == invalid_data_point_id_set:  # All invalid data points
                if len(invalid_data_point_id_set) == 1:
                    invalid_data_point_id = list(invalid_data_point_id_set)[0]
                    raise BadRequestException(
                        f"The specified data point ID does not exist: {invalid_data_point_id}. "
                        f"If you were trying to specify a data set ID, use: data_points='{invalid_data_point_id}'"
                    )
                else:
                    raise BadRequestException("All of the specified data point IDs do not exist.")
            elif len(invalid_data_point_id_set) >= 1:
                invalid_data_points = ", ".join(list(invalid_data_point_id_set))
                raise BadRequestException(f"The specified data point IDs do not exist: {invalid_data_points} or the datapoint settings specified are not valid.")

        elif self.data_point_settings is not None and not self.data_point_settings.empty:
            data_point_list = self._convert_data_point_data_frame(self.data_point_settings)

        elif self.data_set_id is not None and len(self.data_set_id.strip()) > 0:
            data_set_content = self._get_data_set_data_points(None)
            self._insert_name_data_point(data_set_content=data_set_content)
            data_points = _data_point_request_builder(data_set_content)
            # displayName field reflects the column names that are shown in direct for user created data sets.
            # It will have the custom datapoint name if the column name was changed by the user otherwise
            # it will have the default datapoint name. We use desktopDisplayName for datapoints where Direct
            # appends a specific string at the end of the datapoints. For eg. For currency data points (OS060),
            # Base Currency is appended at the end of the datapoint name. In the below logic, the order or preference
            # is desktopDisplayName -> displayName -> defaultDatapointName

            data_point_list = []
            for data_set, data_point in zip(data_set_content, data_points):
                alias_exists = "alias" in data_set and data_set["alias"] == data_point["alias"]
                desktop_display_name_exists = "desktopDisplayName" in data_set and data_set["desktopDisplayName"] is not None
                datapoint_name_exists = "datapointName" in data_set
                if alias_exists and desktop_display_name_exists:
                    display_value = data_set["general"]["desktopDisplayName"]
                elif display_name is True:
                    display_value = data_set["general"]["displayName"]
                else:
                    if datapoint_name_exists:
                        display_value = data_set["datapointName"]  # For user-created data sets
                    else:
                        display_value = data_set["name"]  # For morningstar data sets

                data_point_entry = {**data_point, **{"displayName": display_value}}
                data_point_list.append(data_point_entry)
            if not data_point_list:
                raise BadRequestException("No available data points in the data set.")

        elif list_id is not None and len(list_id.strip()) > 0:
            data_set_content = self._get_binding_data_set_for_user_object(user_object_id=list_id, id_type="LIST")
            self._insert_name_data_point(data_set_content=data_set_content)
            data_point_list = _data_point_request_builder(data_set_content)

        elif search_criteria_id is not None and len(search_criteria_id.strip()) > 0:
            data_set_content = self._get_binding_data_set_for_user_object(user_object_id=search_criteria_id, id_type="SCREEN")
            self._insert_name_data_point(data_set_content=data_set_content)
            data_point_list = _data_point_request_builder(data_set_content)

        if not data_point_list:
            data_point_list = self._get_default_data_point_list()

        result = pd.DataFrame(data_point_list)
        result = result.where(result.notnull(), None)
        return result

    def _insert_name_data_point(self, data_set_content: list) -> None:
        if data_set_content and all(x.get("datapointId") != "OS01W" for x in data_set_content):
            data_set_content.insert(
                0,
                _get_data_point_details([{"datapointId": "OS01W", "isTsdp": False}])[0],
            )

    def _get_binding_data_set_for_user_object(self, user_object_id: str, id_type: str) -> list:
        """
        Get binding data set data points for a investment list or search criteria.
        :param user_object_id: The ID of a saved investment list or search criteria from Direct.
        :param id_type: "LIST" or "SCREEN"
        """
        if user_object_id is not None and len(user_object_id.strip()) > 0 and id_type is not None and len(id_type.strip()) > 0:
            dataset_resp = _get_binding_data_set(user_object_id=user_object_id, id_type=id_type)
            dataset_content = dataset_resp.get("content", []) if dataset_resp else list()
            return self._get_visible_data_points(dataset_content=dataset_content)
        return list()

    def _get_visible_data_points(self, dataset_content: list) -> list:
        available_data_points = []
        if dataset_content:
            for dp in dataset_content:
                desktop_attributions = dp.get("desktopAttributions", None)
                hide = desktop_attributions.get("hide", "0") if desktop_attributions else 0
                if hide != "1":
                    available_data_points.append(dp)
        return available_data_points

    def _validate_data_point_ids_and_alias(self) -> Optional[None]:
        data_point_df = None
        if self.data_point_ids:
            data_point_df = pd.DataFrame(self.data_point_ids)

        elif self.data_point_settings is not None and not self.data_point_settings.empty:
            data_point_df = self.data_point_settings

        if data_point_df is not None and not data_point_df.empty:
            if (
                "datapointId" not in data_point_df.columns
                or "alias" not in data_point_df.columns
                or len(data_point_df[(data_point_df["datapointId"] == "") | (data_point_df["datapointId"].isnull())]) > 0
                or len(data_point_df[(data_point_df["alias"] == "") | (data_point_df["alias"].isnull())]) > 0
            ):
                raise BadRequestException(BAD_REQUEST_ERROR_NO_DATA_POINT_ID_OR_ALIAS)
            if data_point_df[["alias"]].duplicated().sum() >= 1:
                raise BadRequestException(BAD_REQUEST_ERROR_ALIAS_DUPLICATED)

    def get_peer_group_data_points(self) -> pd.DataFrame:
        data_point_df = None
        if self.data_point_ids:
            data_point_df = pd.DataFrame(self.data_point_ids)

        elif self.data_point_settings is not None and not self.data_point_settings.empty:
            data_point_df = self.data_point_settings

        if data_point_df is not None and not data_point_df.empty:
            data_point_df = data_point_df.where(data_point_df.notnull(), None)
            if len(data_point_df[~data_point_df["datapointId"].str.isdigit()]) > 0:
                raise BadRequestException(BAD_REQUEST_ERROR_INCLUDE_NON_CALCULATION_DATA_POINTS)
            if len(data_point_df[["datapointId"]].drop_duplicates()) > 1:
                raise BadRequestException(BAD_REQUEST_ERROR_INCLUDE_DIFF_CALCULATION_DATA_POINTS)

            self._handle_special_data_point_columns(data_point_df)
            data_point_list = self._convert_settings_to_builder(data_point_df)

            if data_point_list:
                result = pd.DataFrame(data_point_list)
                result = result.where(result.notnull(), None)
                return result

        raise BadRequestException(BAD_REQUEST_ERROR_NO_DATA_POINTS)

    def _convert_settings_to_builder(self, settings_data_frame: pd.DataFrame) -> list:
        request_builder = []
        if "isTsdp" not in settings_data_frame.columns:
            settings_data_frame["isTsdp"] = None
        settings_data_frame = settings_data_frame.where(settings_data_frame.notnull(), None)
        default_settings = _get_data_point_details(settings_data_frame[["datapointId", "isTsdp"]].to_dict(orient="records"))
        if default_settings and isinstance(default_settings, list):
            settings = self._replace_value(default_settings, settings_data_frame)
            request_builder = _data_point_request_builder(settings)
            for data_point in request_builder:
                if data_point.get("isTsdp") == True and "frequency" not in data_point:
                    data_point["frequency"] = "m"
        return request_builder

    def _handle_special_data_point_columns(self, settings_data_frame: pd.DataFrame) -> None:
        """
        For values in the columns "windowType", "windowSize", "stepSize" in the settings dataframe, replace empty values with None
        """
        convert_columns = ["windowType", "windowSize", "stepSize"]
        column_list = settings_data_frame.columns.tolist()
        convert_exist_columns = [x for x in convert_columns if x in column_list]
        settings_data_frame[convert_exist_columns] = settings_data_frame[convert_exist_columns].replace(r"^\s*$", None, regex=True)
        settings_data_frame[convert_exist_columns] = settings_data_frame[convert_exist_columns].astype(float).astype(pd.Int64Dtype()).fillna(-1)

    def _convert_data_point_data_frame(self, settings_data_frame: pd.DataFrame) -> list:
        if "OS01W" not in settings_data_frame["datapointId"].tolist():
            name_data_point = pd.DataFrame([{"datapointId": "OS01W", "isTsdp": False}])
            settings_data_frame = pd.concat([name_data_point, settings_data_frame], axis=0)

        settings_data_frame = settings_data_frame.apply(self._set_alias, axis=1)
        settings_data_frame = settings_data_frame.where(settings_data_frame.notnull(), None)

        # separate data points
        asset_flow_data_points = settings_data_frame[settings_data_frame["datapointId"].isin(ALL_ASSET_FLOW_DATA_POINTS)].reset_index().drop(["index"], axis=1)
        normal_data_points = settings_data_frame[~settings_data_frame["datapointId"].isin(ALL_ASSET_FLOW_DATA_POINTS)].reset_index().drop(["index"], axis=1)
        alias_data_point_dict = dict()
        if not asset_flow_data_points.empty:
            default_settings = _get_asset_flow_data_points_by_ids(asset_flow_data_points["datapointId"].tolist())
            if default_settings:
                asset_flow_list = self._replace_asset_flow_settings(default_settings, asset_flow_data_points)
                alias_data_point_dict.update({x.get("alias"): x for x in asset_flow_list})

        if not normal_data_points.empty:
            self._handle_special_data_point_columns(normal_data_points)
            normal_list = self._convert_settings_to_builder(normal_data_points)
            normal_list = normal_list if normal_list else list()
            self._get_no_map_data_points_request_builder(normal_list, normal_data_points)
            alias_data_point_dict.update({x.get("alias"): x for x in normal_list})

        data_point_list = []
        for _, item in settings_data_frame.iterrows():
            target = alias_data_point_dict.get(item["alias"])
            if target:
                display_name = item.get("displayName")
                if display_name is not None:
                    target["displayName"] = str(display_name)
                data_point_list.append(target)
            target = alias_data_point_dict.get(item["alias"] + "_2")
            if target:
                display_name = item.get("displayName")
                if display_name is not None:
                    target["displayName"] = str(display_name)
                data_point_list.append(target)

        return data_point_list

    def _get_data_set_data_points(self, data_point_alias: Optional[list] = None) -> List[Any]:
        if self.data_set_id is not None and re.match(r"^\d{4}-\d{4}$", self.data_set_id, re.M) is not None:  # morningstar data view
            data_set_resp = _get_defaultview_data_points(self.data_set_id)
            if not data_set_resp:
                raise BadRequestException("Data set not Found.")
            data_set_content: List[Any] = data_set_resp.get("content", [])
            if data_set_content:
                for data_point in data_set_content:
                    data_point["datapointId"] = data_point["id"]
            data_set_content = [{**item, **{"displayName": item["general"]["displayName"]}} for item in data_set_content]
            return data_set_content
        else:  # user created data set
            data_set_resp = _get_user_data_set(self.data_set_id)
            if not data_set_resp:
                raise BadRequestException("Data set not Found.")
            data_set_content = data_set_resp.get("content", [])
            sub_data_set_content = [x for x in data_set_content if x.get("alias", "") in data_point_alias] if data_point_alias else data_set_content

            # Add the displayName, desktopDisplayName attribute for each data point in the dataset accordingly
            sub_data_set_content = [
                {
                    **item,
                    **{
                        "displayName": item["general"]["displayName"],
                        "desktopDisplayName": item["general"]["desktopDisplayName"]
                        if "calcCurType" in item["desktopAttributions"]
                        and item["desktopAttributions"]["calcCurType"] != ""
                        and item["desktopAttributions"]["calcCurType"] is not None
                        else None,
                    },
                }
                for item in sub_data_set_content
            ]
            return sub_data_set_content

    def _replace_asset_flow_settings(
        self,
        default_settings: List[Dict[str, Any]],
        asset_flow_data_points: pd.DataFrame,
    ) -> list:
        data_point_list: List[Any] = []
        if asset_flow_data_points is None or asset_flow_data_points.empty:
            return data_point_list
        data_point_id_default_settings_dict = {x.get("datapointId"): x for x in default_settings}
        for _, data_point_setting in asset_flow_data_points.iterrows():
            data_point_id = data_point_setting["datapointId"]
            default_setting = data_point_id_default_settings_dict.get(data_point_id, None)
            if not default_setting:
                continue
            default_setting_copy = copy.deepcopy(default_setting)
            for key, value in data_point_setting.items():
                if value is None:
                    continue
                if key in default_setting_copy.keys() or key == "marketId":
                    default_setting_copy[key] = value
            data_point_list.append(default_setting_copy)

        return data_point_list

    def _set_alias(self, settings: dict) -> dict:
        # Normally, aliases are not specified by users when specifying data point settings
        # for methods like get_investment_data() as they are auto-generated by the code below.
        # Other parts of the code include the alias in the request for certain DO API calls so it can
        # map data points in the API request to data point values in the API response.
        # The only time we need to specify aliases in data point settings is when we do unit tests to
        # ensure we use the same set of UUIDs on every test run.
        if "alias" not in settings or not _is_uuid(str(settings["alias"])):
            settings["alias"] = str(uuid.uuid4())
        return settings

    def _replace_value(self, default_settings: list, data_point_settings: pd.DataFrame) -> list:
        data_point_list = []
        for _, data_point_setting in data_point_settings.iterrows():
            data_point_id = data_point_setting["datapointId"]
            is_ts_dp = data_point_setting.get("isTsdp", None)
            if isinstance(is_ts_dp, bool):
                target = self._filter_data_point(data_point_id, is_ts_dp, default_settings)
                if target is not None:
                    data_point_list.append(self._copy_and_replace_value(target, data_point_setting))
            else:
                data_points = list()
                target_ts = self._filter_data_point(data_point_id, True, default_settings)
                target_current = self._filter_data_point(data_point_id, False, default_settings)
                if target_ts is not None and target_current is not None:
                    """
                    If target_ts is not None and target_current is not None, the datapoint is a time series datapoint.
                    In this scenario, if startDate and endDate are passed, only get data for target_ts.
                    """
                    if self._is_start_end_date_available(data_point_setting):
                        target_current = None

                if target_ts is not None:
                    data_points.append(self._copy_and_replace_value(target_ts, data_point_setting))

                if target_current is not None:
                    data_points.append(self._copy_and_replace_value(target_current, data_point_setting))
                if len(data_points) == 2:
                    data_points[1]["alias"] = data_points[1]["alias"] + "_2"
                if data_points:
                    data_point_list.extend(data_points)
        return data_point_list

    def _is_start_end_date_available(self, data_point_setting: Series) -> bool:
        start_date = data_point_setting.get("startDate", None)
        end_date = data_point_setting.get("endDate", None)
        if start_date is not None and len(start_date.strip()) > 0 and end_date is not None and len(end_date.strip()) > 0:
            return True
        return False

    def _filter_data_point(self, data_point_id: str, is_ts: bool, default_settings: list) -> Optional[dict]:
        target_list = list(
            filter(
                lambda x: x.get("datapointId", "") == data_point_id and x.get("isTsdp", "") == is_ts,
                default_settings,
            )
        )
        return target_list[0] if len(target_list) > 0 else None

    def _copy_and_replace_value(self, target: dict, data_point_setting: pd.Series) -> dict:
        target_copy = copy.deepcopy(target)
        for key, value in data_point_setting.items():
            if value is None:
                continue
            if key in target_copy.keys() or key == "alias":
                target_copy[key] = value
            else:
                for name_key in [
                    "general",
                    "source",
                    "calculation",
                    "desktopAttributions",
                ]:
                    if name_key not in target_copy:
                        target_copy[name_key] = {}
                    target_copy[name_key][key] = value

        return target_copy

    def _get_no_map_data_points_request_builder(self, request_builder_list: list, df: pd.DataFrame) -> None:
        no_map_data_points = [
            "DC09A",
            "TE190",
            "TE158",
            "TE157",
            "TE156",
            "TE164",
            "TE163",
            "TE162",
            "TE161",
            "TE218",
        ]
        for row in df.itertuples():
            data_point_id = getattr(row, "datapointId")
            if data_point_id in no_map_data_points:
                no_map_data = {
                    "alias": getattr(row, "alias"),
                    "annualDays": "0",
                    "datapointId": data_point_id,
                    "isTsdp": getattr(row, "isTsdp"),
                    "requireContinueData": True,
                    "sourceId": getattr(row, "datapointId"),
                }

                if no_map_data["isTsdp"] is True:
                    for name_column in ["frequency", "startDate", "endDate"]:
                        if getattr(row, name_column):
                            no_map_data[name_column] = getattr(row, name_column)

                request_builder_list.append(no_map_data)

    def _convert_morningstar_data_points_object(self, data_points: list) -> pd.DataFrame:
        """
        Converts a list of morningstar_data_points DataPoints or TimeSeriesDataPoints objects into a
        data points settings DataFrame object.
        """
        formatted_points = []
        for point in data_points:
            if type(point) != dict:
                formatted_points.append(self._convert_data_point(point))
            else:
                formatted_points.append(point)
        return pd.DataFrame(formatted_points)

    def _convert_data_point(self, dp: dict) -> dict:
        """Converts a single data point.

        Creates a single data point in the format for get_investment_data.

        Args:
            dp: A data point

        Returns:
            Vals: A dictionary of the needed settings for get_investment_data function
        """

        vals = {}
        for k, v in dp.__dict__.items():
            if k == "name":
                pass
            elif k == "datapointId":
                vals[k] = v
            elif k == "sourceData":
                vals["sourceId"] = str(v)
            elif isinstance(v, datetime):
                vals[k] = (str(v))[0:10]
            else:
                vals[k] = str(v)

        vals["datapointName"] = dp.name  # type: ignore
        dp_module_name = inspect.getmodule(type(dp)).__name__  # type: ignore
        if "_not_tsdp" in dp_module_name:
            vals["isTsdp"] = False
        else:
            vals["isTsdp"] = True

        return vals

    def _is_list_of_dicts(self, data_points_list: list) -> bool:
        return all(isinstance(elem, dict) for elem in data_points_list)

    def _get_default_data_point_list(self) -> list:
        default_settings_data_frame = pd.DataFrame(DefaultDataPoints)
        default_data_point_list = self._convert_data_point_data_frame(default_settings_data_frame)
        return default_data_point_list



import simplejson as json
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Union

import requests

# NOTE! We are not able to verify the SSL cert on Direct Search API since
#       the VPC endpoint domain name is not on the server's SSL cert.
#       So with a heavy heart, we disable the InsecureRequestWarning
from urllib3.exceptions import InsecureRequestWarning

from .._base import _logger
from . import _error_messages
from ._config import _Config
from ._exceptions import AccessDeniedError, BadRequestException, ClientError, ForbiddenError, InternalServerError, NetworkExceptionError, TimeoutError
from ._core_api import make_request

_config = _Config()

requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)


class APIBackend(ABC):
    """
    Abstract class for all API requests
    """

    def do_get_request(self, url: str) -> Any:
        """
        Makes a GET request
        """
        return self._make_request(url, "GET")

    def do_post_request(self, url: str, data: Optional[Union[Dict[Any, Any], str]] = None) -> Any:
        """
        Makes a POST request
        """
        return self._make_request(url, "POST", data)

    def do_put_request(self, url: str, data: Optional[Union[Dict[Any, Any], str]] = None) -> Any:
        """
        Makes a PUT request
        """
        return self._make_request(url, "PUT", data)

    def get_headers(self) -> Dict[str, str]:
        return _config.get_headers()

    def _make_request(self, url: str, method: str, data: Optional[Union[Dict[Any, Any], str]] = None) -> Any:
        """
        Request and handle OK (status_code = 200) and network error responses.
        """
        _logger.info(f"Requesting: {method} {url}")
        try:
            res = make_request(method, url, headers=self.get_headers(), verify=False, data=data)
            res.raise_for_status()
            return res.json()
        except requests.ConnectionError as e:
            raise NetworkExceptionError from None
        except requests.HTTPError as e:
            self._handle_custom_http_errors(res)
            self._handle_default_http_errors(res)

    def _handle_default_http_errors(self, res: Any) -> Any:
        """
        Handle HTTP errors with standard messages
        """
        try:
            response_message = res.json()["message"]
        except KeyError:
            response_message = res.json()
            _logger.debug("DO Request Error")
        if res.status_code == 401:
            _logger.debug(f"Access Denied Error: {res.status_code} {response_message}")
            raise AccessDeniedError from None
        elif res.status_code == 403:
            _logger.debug(f"Forbidden Error: {res.status_code} {response_message}")
            raise ForbiddenError from None
        elif res.status_code == 408:
            _logger.debug(f"Timeout Error: {res.status_code} {response_message}")
            raise TimeoutError from None
        elif res.status_code == 400:
            _logger.debug(f"Bad Request: {res.status_code} {response_message}")
            raise BadRequestException(_error_messages.BAD_REQUEST_ERROR) from None
        elif res.status_code > 400 and res.status_code < 500:
            _logger.debug(f"Client Error: {res.status_code} {response_message}")
            raise ClientError(_error_messages.CLIENT_ERROR) from None
        else:
            _logger.debug(f"Internal Server Error: {res.status_code} {response_message}")
            raise InternalServerError(_error_messages.INTERNAL_SERVER_ERROR) from None

    @abstractmethod
    def _handle_custom_http_errors(self, res: Any) -> None:
        """
        Method to handle special HTTP errors with custom error messages
        """
        pass



from pandas import DataFrame
from abc import ABC, abstractmethod


class Table(ABC):
    def __init__(self, name: str, df: DataFrame) -> None:
        self.name = name
        self.df = self._validate_column_names(df)

    @abstractmethod
    def _validate_column_names(self, df: DataFrame) -> None:
        """
        Validate that column names in Data Frame do not contain reserved words and conform to Redshift's requirements:
        https://docs.aws.amazon.com/redshift/latest/dg/r_names.html

        Args:
            DataFrame
        Returns:
            The validated DataFrame
        """
        pass

    @abstractmethod
    def get_create_query(self) -> None:
        pass

    @abstractmethod
    def get_insert_query(self) -> None:
        pass



import enum
from enum import Enum
from typing import Any
from ._exceptions import BadRequestException


class DatabaseCD(str, Enum):
    user = ("UDP",)
    firm = ("CDP",)


class RaiseEnum(enum.EnumMeta):
    def __getitem__(cls, name: str) -> Any:
        try:
            return super().__getitem__(name)
        except KeyError:
            options = ", ".join(cls._member_map_.keys())
            raise BadRequestException(f"Please choose one of the following valid options: '{options}'.\n" f"'{name}' is an invalid choice.\n") from None


class ErrorMessages(str, Enum):
    date_format_error = "Please input the date in yyyy-MM-dd format."
    start_date_require_error = "Please specify start_date to proceed with your query."
    no_data_point_msg = "There is no data point available."
    start_end_date_error = "Please specify start_date and end_date to proceed with your query."
    frequency_error_msg = "Please specify frequency to proceed with your query."
    currency_error_msg = "Please specify currency to proceed with your query."


class DryRunResults:
    estimated_cells_used: int
    daily_cells_remaining_before: int
    daily_cells_remaining_after: int
    daily_cell_limit: int

    def __init__(self, estimated_cells_used: int, daily_cells_remaining_before: int, daily_cell_limit: int) -> None:
        self.estimated_cells_used = estimated_cells_used
        self.daily_cells_remaining_before = daily_cells_remaining_before
        self.daily_cell_limit = daily_cell_limit
        self.daily_cells_remaining_after = daily_cells_remaining_before - estimated_cells_used



from . import direct
from . import lookup
from . import datalake
from . import utils
from .datalake.base import query
from .datalake._data_objects import CSVFile, TempTable
from ._version import __version__



import re

from datetime import datetime
from typing import Union, Callable

from .._exceptions import ValueErrorException


TypeValue = Union[str, int, float, None]


class ValidCustomDataPointTypes:
    def __init__(self, data_point_definition: dict):
        function_name = re.sub(r"(?<!^)(?=[A-Z])", "_", data_point_definition["columnType"]).lower()
        self.is_valid: Callable = getattr(ValidCustomDataPointTypes, function_name, self.raise_value_error)

    def raise_value_error(self, val: TypeValue) -> None:
        raise ValueErrorException("Column type is not supported.")

    @staticmethod
    def extended_text(val: TypeValue) -> TypeValue:
        if isinstance(val, str) and len(val) <= 250:
            return val
        raise ValueErrorException("Only allows a string data type with 250 characters or less")

    @staticmethod
    def numeric(val: TypeValue) -> TypeValue:
        if isinstance(val, int) or isinstance(val, float):
            return val
        raise ValueErrorException("Only allows numbers")

    @staticmethod
    def free_text(val: TypeValue) -> TypeValue:
        if isinstance(val, str) and len(val) <= 50:
            return val
        raise ValueErrorException("Only allows a string data type with 50 characters or less")

    @staticmethod
    def date(val: TypeValue) -> TypeValue:
        format_Data = "%Y-%m-%d"
        if isinstance(val, str):
            datetime.strptime(val, format_Data)
            return val
        raise ValueErrorException("Only allows Date data type with the format yyyy-mm-dd")

    @staticmethod
    def category(val: TypeValue) -> TypeValue:
        return ValidCustomDataPointTypes.free_text(val)

    @staticmethod
    def indicator(val: TypeValue) -> TypeValue:
        return ValidCustomDataPointTypes.numeric(val)

    @staticmethod
    def t_s_return(val: TypeValue) -> TypeValue:
        return ValidCustomDataPointTypes.numeric(val)

    @staticmethod
    def t_s_price(val: TypeValue) -> TypeValue:
        return ValidCustomDataPointTypes.numeric(val)

    @staticmethod
    def t_s_numeric(val: TypeValue) -> TypeValue:
        return ValidCustomDataPointTypes.numeric(val)

    @staticmethod
    def t_s_free_text(val: TypeValue) -> TypeValue:
        return ValidCustomDataPointTypes.free_text(val)



import simplejson as json
import warnings
import requests

from typing import Any, Dict, List, Union, Optional
from urllib.parse import quote

from pandas import DataFrame

from ..._base import _logger
from .. import _decorator, _error_messages, investment
from .._base_api import APIBackend
from .._config import _Config
from .._exceptions import BadRequestException, ResourceNotFoundError, NetworkExceptionError
from .._utils import _reduce_list_data
from . import _utils

_config = _Config()


class SearchCriteriaAPIBackend(APIBackend):
    """
    Subclass to call the Search Criteria API and handle any HTTP errors that occur.
    """

    def __init__(self) -> None:
        super().__init__()

    def _handle_custom_http_errors(self, res: Any) -> Any:
        """
        Handle HTTP errors with custom error messages
        """
        response_message = res.json()["message"]
        if res.status_code == 404:
            _logger.debug(f"Resource Not Found Error: {res.status_code} {response_message}")
            raise ResourceNotFoundError(_error_messages.RESOURCE_NOT_FOUND_ERROR_SEARCH_CRITERIA) from None
        else:
            pass


_search_criteria_api = SearchCriteriaAPIBackend()


@_decorator.not_null
@_decorator.typechecked
def get_search_results(search_criteria_id: str) -> DataFrame:
    """Returns all investments matching the specified search criteria.

    Args:
        search_criteria_id (:obj:`str`): Unique identifier of a saved search criteria in Morningstar Direct, e.g., "9009". Use `get_search_criteria <#morningstar_data.direct.user_items.get_search_criteria>`_ to discover possible values.

    Returns:
        DataFrame: A DataFrame object with investments that match the search criteria. DataFrame columns include:

        * secid
        * masterportfolioid
        * tradingsymbol
        * name
        * securitytype
        * exchangeid
        * category

    Examples:
        Get investments that match the given search criteria.

    ::

        import morningstar_data as md

        df = md.direct.user_items.get_search_results(search_criteria_id="4237053") # Replace with a valid search criteria id
        df

    :Output:
        ==========  =================  =============  =======================  ============  ==========  ================
        secid       masterportfolioid  tradingsymbol  name                     securitytype  exchangeid  category
        ==========  =================  =============  =======================  ============  ==========  ================
        FOUSA06JNH  210311             AAAAX          DWS RREEF Real Assets A  FO            EXXNAS      World Allocation
        ...
        ==========  =================  =============  =======================  ============  ==========  ================

    Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    try:
        search_df = DataFrame(columns=_utils.investments_list_default_columns)
        _logger.debug(f"Getting Direct SearchCriteria: {search_criteria_id}")

        search_criteria_id_encode = quote(search_criteria_id, "utf-8")
        url = f"{_config.searches_service_url()}v1/searches/{search_criteria_id_encode}"

        response_json = _search_criteria_api.do_get_request(url)
        search = response_json["content"]

        _logger.debug(f"Executing Direct SearchCriteria: {search_criteria_id}")
        investments = execute_search(search)

        # [{"id": "1;a"}, {"id": "2;x"}] => "'1','2'"
        ids = list(map(lambda x: f"{x['id']}", investments))

        if not ids:
            _logger.debug(f"No ids returned in Direct Searches API response, creating empty DataFrame.")
            search_df = DataFrame(columns=_utils.investments_list_default_columns)
        else:
            search_df = investment.get_investment_data(investments=ids, data_points=_utils.investments_list_data_points)
            search_df = search_df.drop(["Id"], axis=1)
            search_df.columns = _utils.investments_list_default_columns

        if search_df.empty:
            _logger.debug(f"Returning empty DataFrame.")
        _logger.debug(f"Returning DataFrame for SearchCriteria: {search_criteria_id}")
        return search_df

    except KeyError as e:
        raise BadRequestException(_error_messages.BAD_REQUEST_ERROR) from None


@_decorator.typechecked
def get_search_criterias() -> DataFrame:
    warnings.warn(
        "The get_search_criterias function is deprecated and will be removed in the next major version. Use get_search_criteria instead",
        FutureWarning,
        stacklevel=2,
    )
    return get_search_criteria()


@_decorator.typechecked
def get_search_criteria() -> DataFrame:
    """Returns all search criteria saved or shared to a user in Morningstar Direct

    Returns:
        DataFrame: A DataFrame object with all search criteria names and identifiers. DataFrame columns include:

        * id
        * name

    :Examples:

    ::

        import morningstar_data as md

        df = md.direct.user_items.get_search_criteria()
        df

    :Output:
        =======  ======
        id       name
        =======  ======
        4175985  sample
        5022284  others
        ...
        =======  ======

    Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    try:
        _logger.debug(f"Getting all Direct SearchCriteria")

        url = f"{_config.searches_service_url()}v1/searches"

        response_json = _search_criteria_api.do_get_request(url)
        props_dict: Dict[str, str] = {"id": "searchId", "name": "name"}
        all_search_criterias = _reduce_list_data(response_json["search"], props_dict)
        return DataFrame(all_search_criterias)
    except Exception as e:
        if isinstance(e, ResourceNotFoundError):
            raise ResourceNotFoundError(_error_messages.RESOURCE_NOT_FOUND_ERROR_NO_SEARCH_CRITERIA) from None
        raise e from None


@_decorator.not_null
@_decorator.typechecked
def get_search_criteria_condition(search_criteria_id: str) -> Any:
    warnings.warn(
        "The get_search_criteria_condition function is deprecated and will be removed in the next major version. Use get_search_criteria_conditions instead",
        FutureWarning,
        stacklevel=2,
    )
    return get_search_criteria_conditions(search_criteria_id)


@_decorator.not_null
@_decorator.typechecked
def get_search_criteria_conditions(search_criteria_id: str) -> Dict[str, Any]:
    """Returns the detailed definition of a saved search criteria.

    Args:
        search_criteria_id (:obj:`str`): Unique identifier of a saved search criteria in Morningstar Direct, e.g., "9009". Use `get_search_criteria <#morningstar_data.direct.user_items.get_search_criteria>`_ to discover possible values.

    Returns:
        Dict:

    :Examples:

    ::

        import morningstar_data as md

        df = md.direct.user_items.get_search_criteria_conditions(search_criteria_id="9009") # Replace with a valid search criteria id
        df

    :Output:

    ::

        {
            "universeId": "FO",
            "subUniverseId": "",
            "subUniverseName": "",
            "securityStatus": "activeonly",
            "useDefinedPrimary": False,
            "criteria": [
                {
                    "relation": "",
                    "field": "OS001",
                    "operator": "=",
                    "value": "AAAAX",
                    "id": "FOUSA06JNH",
                    "name": "AAAAX"
                }
            ]
        }

    Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        ForbiddenError: Raised when the user does not have permission to access the requested resource.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """

    try:
        search_criteria_id_encode = quote(search_criteria_id, "utf-8")
        url = f"{_config.searches_service_url()}v1/searches/{search_criteria_id_encode}"
        response_json = _search_criteria_api.do_get_request(url)
        condition: Dict[str, Any] = response_json.get("content", {})
        return condition
    except Exception as e:
        if isinstance(e, ResourceNotFoundError):
            raise ResourceNotFoundError(_error_messages.RESOURCE_NOT_FOUND_ERROR_SEARCH_CRITERIA_ID.format(search_criteria_id)) from None
        raise e from None


def execute_search(search: dict) -> List[Dict[str, Any]]:
    request_body = json.dumps({"search": search}, ignore_nan=True)
    url = f"{_config.securitydata_service_url()}v1/searchresults"

    response_json: Dict[Any, Any] = _search_criteria_api.do_post_request(url, data=request_body)
    investments: List[Dict[str, Any]] = response_json.get("investments", [])
    if "investments" not in response_json:
        _logger.error(f"No investments in search results: {response_json}")
        raise ResourceNotFoundError(_error_messages.RESOURCE_NOT_FOUND_ERROR_SEARCH_CRITERIA_NO_INVESTMENTS) from None
    return investments



from __future__ import annotations
import datetime
import simplejson as json
from enum import Enum
from sre_constants import ANY
from typing import Any, Dict

from pandas import DataFrame

from ._api import _direct_api_request, _config


def _annual_return_year(year: int = 0) -> str:
    current_date = datetime.date.today()
    new_year = current_date.year - year
    return str(new_year)


def _update_annual_return_alias(data_point_id: str) -> str:
    maps = {"AR002": 1, "AR003": 2, "AR004": 3, "AR005": 4, "AR006": 5, "AR007": 6, "AR008": 7, "AR009": 8, "AR00A": 9, "AR00B": 10, "AR00C": 11}
    map_value = maps.get(data_point_id, None)
    if map_value is not None:
        return _annual_return_year(map_value)
    else:
        return ""


def _get_data_sets() -> DataFrame:
    url = f"{_config.asset_service_url()}/portfolio/datasets"
    response_json = _direct_api_request("get", url)
    return DataFrame(response_json)


def _get_data_set_with_id(data_set_id: str) -> Any:
    try:
        if data_set_id is not None:
            url = f"{_config.asset_service_url()}portfolio/dataset/{data_set_id}"
            response_json = _direct_api_request("get", url)
    except Exception:
        response_json = {"datapoints": []}
    response_json = _update_calendar_year_data_points_name(response_json)
    return response_json


def _update_calendar_year_data_points_name(response_json: Any) -> Any:
    data_set_id = response_json.get("id", None)
    if data_set_id is not None and PortfolioDataSet.get(str(data_set_id)) == PortfolioDataSet.Returns_Calendar_Year:
        data_points = response_json.get("datapoints", [])
        for data_point in data_points:
            new_alias = f'{data_point.get("alias", "").strip()} {_update_annual_return_alias(data_point.get("datapointId"))}'
            data_point["alias"] = new_alias
    return response_json


def _get_data_points_default_settings(data_points: list) -> Dict[str, dict]:
    url = f"{_config.asset_service_url()}portfolio/datapoints"
    post_body = list(map(lambda x: {"datapointId": x}, data_points))
    response_json = _direct_api_request("post", url, json.dumps(post_body, ignore_nan=True))
    defined_data_points = {x.get("datapointId"): x for x in response_json}
    return defined_data_points


class DataSetType(Enum):
    Current = 1
    TimeSeries = 2
    Current_Or_TimeSeries = 3
    Custom_Calculation = 4
    No_Map = 0


class PortfolioDataSet(Enum):
    def __init__(self, data_set_id: str, data_set_type: DataSetType) -> None:
        self.data_set_type = data_set_type
        self.data_set_id = data_set_id

    @classmethod
    def get(cls, data_set_id: str) -> Any:
        for member in cls._member_map_.values():
            if member._value_[0] == data_set_id:
                return member
        return PortfolioDataSet["No_Map"]

    @classmethod
    def get_data_set_type(cls, data_set_id: str) -> Any:
        for member in cls._member_map_.values():
            if member._value_[0] == data_set_id:
                return member._value_[1]
        return DataSetType["No_Map"]

    No_Map = ("no_map", DataSetType.No_Map)
    Asset_Allocation = ("99", DataSetType.Current_Or_TimeSeries)
    Custom_Calculation = ("8", DataSetType.Custom_Calculation)
    Daily_Gross_Return_Index = ("7", DataSetType.TimeSeries)
    Daily_Return_Index = ("6", DataSetType.TimeSeries)
    Equity_Country_Exposure = ("15", DataSetType.Current_Or_TimeSeries)
    Equity_Market_Capitalization = ("25", DataSetType.Current_Or_TimeSeries)
    Equity_Port_Stats_Long = ("27", DataSetType.Current_Or_TimeSeries)
    Equity_Port_Stats_Short = ("28", DataSetType.Current_Or_TimeSeries)
    Equity_Regional_Exposure = ("21", DataSetType.Current_Or_TimeSeries)
    Equity_Sector_Exposure = ("17", DataSetType.Current_Or_TimeSeries)
    Equity_Style_Analysis = ("24", DataSetType.Current_Or_TimeSeries)
    Equity_Style_Capitalization = ("22", DataSetType.Current_Or_TimeSeries)
    Equity_Style_Valuation = ("23", DataSetType.Current_Or_TimeSeries)
    Equity_Type = ("19", DataSetType.TimeSeries)
    ESG_Carbon = ("12", DataSetType.Current_Or_TimeSeries)
    ESG_Product_Involvement = ("13", DataSetType.Current_Or_TimeSeries)
    ESG_Sustainability = ("14", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_ABS_Collateral_Type_Detail = ("34", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Bond_Insurer = ("71", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Call_Type = ("35", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_CMO_Issuer_Type = ("36", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_CMO_Payment_Type = ("37", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_CMO_Tranche_Type = ("38", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Convertible_Type = ("43", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Country_Exposure = ("29", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Coupon_Type = ("44", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Credit_Enhancement_Type = ("67", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Credit_Grd_And_Credit_Rtg_Breakdown_Detail = ("77", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_CreditQualityBreakdown = ("68", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Credit_Rtg = ("76", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Current_Yield_And_Current_Yield_Breakdown_Detail = ("85", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Default_Reason = ("61", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Default_Type = ("62", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_DTW_Breakdown_Super = ("82", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_DTW_And_DTW_Breakdown_Detail = ("75", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Duration_Effective = ("39", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Duration_Modified_to_maturity = ("40", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Eff_Convxty_breakdown_Super = ("84", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_EFF_Convxty_And_Convxty_Breakdown_Detail = ("73", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Eff_Dua_Breakdown_Super = ("78", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Effective_Maturity = ("41", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Eff_Mty_Breakdown_Super = ("80", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Geographic_Exposure = ("93", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Guarantee_Type = ("63", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Guarantor_Type = ("64", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Interest_Rate_Type_Detail = ("45", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Issuer_Type = ("46", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Issuer_Type_Breakdown_Super = ("90", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_LOC_Type = ("65", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Maturity_Type = ("47", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_MBS_Agency = ("48", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Mbs_Agency_Breakdown_Super = ("87", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_MBS_Collateral_Type = ("49", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Mod_Dur_Breakdown_Super = ("79", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Morningstar_Sectors = ("94", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Municipal_Security_Type_Breakdown = ("92", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Municipal_Security_Type_Breakdown_Super = ("88", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Next_Call_And_Next_Call_Breakdown_Detail = ("86", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_OAS_Breakdown_Super = ("83", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_OAS_And_OAS_Breakdown_Detail = ("72", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_PIK_Type = ("50", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Portfolio_Statistics = ("33", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Pre_refunded_ETM_Type = ("66", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Primary_Sector_Breakdown = ("32", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Put_Type = ("51", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Put_Type_Detail = ("52", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Secondary_Sector_Breakdown = ("30", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_SEC_Registration_Type = ("53", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Secured_by_Collateral_Type = ("54", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Securitization_Collateral_Type = ("69", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Securitized_Mortgage_Type = ("55", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Security_Rank = ("56", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Security_Rank_Breakdown_Super = ("91", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Sinking_Fund_Type = ("57", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_State_Taxation = ("58", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Structured_Securitized_Type = ("59", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Super_Sector_Breakdown = ("31", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Use_of_Proceeds = ("70", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Use_Of_Proceeds_Breakdown_Super = ("89", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_US_Federal_Taxation = ("60", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_Yield_to_Maturity_YTM = ("42", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_YTM_Breakdown_Super = ("81", DataSetType.Current_Or_TimeSeries)
    Fixed_Income_YTW_And_YTW_Breakdown_Detail = ("74", DataSetType.Current_Or_TimeSeries)
    GICS_Equity_Industry = ("20", DataSetType.Current_Or_TimeSeries)
    GICS_Equity_Sector_Exposure = ("18", DataSetType.Current_Or_TimeSeries)
    MSCI_Equity_Country_Exposure = ("16", DataSetType.Current_Or_TimeSeries)
    Post_tax_Returns_Month_end = ("9", DataSetType.Current_Or_TimeSeries)
    Regional_Asset_Class_Market_Cap = ("95", DataSetType.Current)
    Representative_Cost = ("96", DataSetType.Current_Or_TimeSeries)
    Returns_Calendar_Year = ("5", DataSetType.Current)
    Returns_Daily = ("2", DataSetType.Current)
    Returns_Month_End = ("3", DataSetType.Current)
    Returns_Quarter_End = ("4", DataSetType.Current)
    Risk_Total_ReturnMonth_End = ("10", DataSetType.Current_Or_TimeSeries)
    Risk_Total_ReturnQtr_End = ("11", DataSetType.Current_Or_TimeSeries)
    Snapshot = ("1", DataSetType.Current)
    South_Africa_Asset_Allocation = ("101", DataSetType.Current_Or_TimeSeries)
    Stock_Intersection = ("26", DataSetType.Current_Or_TimeSeries)
    Surveyed_Asset_Allocation_AUS = ("97", DataSetType.Current_Or_TimeSeries)
    Surveyed_Asset_Allocation_CAN = ("98", DataSetType.Current_Or_TimeSeries)
    Surveyed_Asset_Allocation_NZ = ("100", DataSetType.Current_Or_TimeSeries)
